-----------------------------------------------------------------------------------------------
--[[ Copyright (c) chckxy (watdametz@gmail.com). All rights reserved

This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.  --]]
-----------------------------------------------------------------------------------------------
 
local Flip = {} 

function Flip:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

function Flip:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Flip"
	local tDependencies = {
		"MarketplaceCommodity", 
		"MarketplaceAuction",
		-- other flip packages:
		"FlipGroups",
		"FlipCommodity",
		"FlipAuction",
		"FlipMarket",
		"FlipCrafting",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

function Flip:OnLoad()
	self.bLoadOk = true

	self.nMinFontSize = 10

	-- groups
	self.groups = Apollo.GetPackage("FlipGroups").tPackage:new({ 
		flip = self,
		nMinFontSize = self.nMinFontSize,
	})

	-- other modules
	self.market = Apollo.GetPackage("FlipMarket").tPackage:new({
		flip = self, 
		groups = self.groups, 
		nMinFontSize = self.nMinFontSize,
	})
	self.craft = Apollo.GetPackage("FlipCrafting").tPackage:new({
		flip = self, 
		groups = self.groups,
		market = self.market,
		nMinFontSize = self.nMinFontSize,
	})

	-- register slash commands
	Apollo.RegisterSlashCommand("flip", "OnFlipOn", self.market)
	Apollo.RegisterSlashCommand("flipmarket", "OnFlipMarketOn", self.market)
	Apollo.RegisterSlashCommand("flipcraft", "OnFlipCraftOn", self.craft)
	Apollo.RegisterSlashCommand("flipgroup", "OnFlipGroupOn", self.groups)

	-- many events such market
	Apollo.RegisterEventHandler("InvokeVendorWindow", "OnInvokeVendorWindow", self.market)	
	Apollo.RegisterEventHandler("CommodityInfoResults", "OnCommodityInfoResults", self.market)
	Apollo.RegisterEventHandler("ItemAuctionSearchResults", "OnItemAuctionSearchResults", self.market)
	Apollo.RegisterEventHandler("ToggleMarketplaceWindow", "OnToggleMarketplaceWindow", self.market)
	Apollo.RegisterEventHandler("ToggleAuctionWindow", "OnToggleAuctionWindow", self.market)
	Apollo.RegisterEventHandler("PostCommodityOrderResult", "RefreshCX", self.market)
end

-- CharacterLevelSave : SaveLeveltype -> Boolean
function Flip:CharacterLevelSave( eType )
	return eType == GameLib.CodeEnumAddonSaveLevel.Character
end

-- OnSave : SaveLevelType -> Save
-- if anything goes wrong, use the previous save as to not lose data
function Flip:OnSave( eType )
	-- save a different list for each character, since listings are separate anyway
	if not self:CharacterLevelSave(eType) then return nil end

	-- if we tried to restore earlier but failed, then abandon ship
	if not self.bLoadOk then return nil end

	local tNewSave = {
		tGroupsSave = (self.groups and self.groups:ToSave()) or self.tPreviousSave.tGroupsSave,
		tMarketSave = (self.market and self.market:ToSave()) or self.tPreviousSave.tMarketSave,
		tCraftSave  = (self.craft  and  self.craft:ToSave()) or self.tPreviousSave.tCraftSave,
		nMinFontSize = self.nMinFontSize or 10,
	}

	return tNewSave
end

-- OnRestore : SaveLevelType Save ->
-- be really sure never screw up the group list
function Flip:OnRestore( eType, tSave )
	-- only use the character level save
	if not self:CharacterLevelSave(eType) then return nil end

	self.bLoadOk = false
	local bLoadOk = true

	-- version with groups
	if tSave.tGroupsSave then
		bLoadOk =             self.groups and self.groups:RestoreSave(tSave.tGroupsSave)
		bLoadOk = bLoadOk and self.market and self.market:RestoreSave(tSave.tMarketSave)
		bLoadOk = bLoadOk and self.craft  and self.craft:RestoreSave(tSave.tCraftSave)
		self:SetMinFontSize(tSave.nMinFontSize or 10)
	end

	-- if we loaded ok, then save this save in case saving don't save so good
	if bLoadOk then self.tPreviousSave = tSave end
	self.bLoadOk = bLoadOk
end

function Flip:GetModules()
	return { MARKET = self.market, GROUPS = self.groups, } --CRAFT = self.craft, }
end

function Flip:OnConfigure()
	self.market:OnConfigure()
end

-- TODO: make an event
function Flip:SetMinFontSize( nSize )
	if nSize ~= 9 and (nSize % 2) ~= 0 then nSize = nSize + 1 end
	self.nMinFontSize = nSize
	self.groups:SetMinFontSize(nSize)
	self.market:SetMinFontSize(nSize)
	self.craft:SetMinFontSize(nSize)
end

---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

local FlipInst = Flip:new()
FlipInst:Init()
