-----------------------------------------------------------------------------------------------
--[[ Copyright (c) chckxy (watdametz@gmail.com). All rights reserved

This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.  --]]
-----------------------------------------------------------------------------------------------

local FlipAuction = {}
local FlipAuctionUpdater = {}
local FlipAuctionListing = {}

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipAuction:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
	return o
end

function FlipAuction:Init()
	-- we have a market
	self.market = self.market
	
	-- we have the groups module
	self.groups = self.groups

	-- headers being displayed
	self.tUpdaters = {}

	self.xmlDoc = XmlDoc.CreateFromFile("Flip.xml")	
	self.wndMain = self.wndMain
	self.wndTabs = self.wndMain:FindChild("TabContainer")
	self.wndSubTab = self.wndMain:FindChild("SubTabContainer")
	self.wndLeft = self.wndMain:FindChild("Left")
	self.wndRight = self.wndMain:FindChild("Right")

	-- timers for (refresh) delay
	self.timerDelay = Apollo.CreateTimer("AuctionTimerDelay", 2, false)
	Apollo.StopTimer("AuctionTimerDelay")
	Apollo.RegisterTimerHandler("AuctionTimerDelay", "RefreshAH", self.market)

	self.timerDirty = Apollo.CreateTimer("AuctionTimerDirty", 2, false)
	Apollo.StopTimer("AuctionTimerDirty")
	Apollo.RegisterTimerHandler("AuctionTimerDirty", "Redraw", self)

	-- our filter is at zero initially
	self.nFilterAmount = 0
	
	-- add the watch button to all the auctionables; got idea from CommodityStats
	local mc = Apollo.GetAddon("MarketplaceAuction")
	local fnOldOnItemAuctionSearchResults = mc.OnItemAuctionSearchResults
	mc.OnItemAuctionSearchResults = function(tMarketplaceAuction, nPage, nTotalResults, tAuctions)
		fnOldOnItemAuctionSearchResults(tMarketplaceAuction, nPage, nTotalResults, tAuctions)
		local wndParent = nil
		if not tMarketplaceAuction.wndMain then return nil end
		local bBuyTab = tMarketplaceAuction.wndMain:FindChild("HeaderBuyBtn"):IsChecked()
		if bBuyTab then
			wndParent = tMarketplaceAuction.wndMain:FindChild("BuyContainer"):FindChild("SearchResultList")
		else
			wndParent = tMarketplaceAuction.wndMain:FindChild("SellContainer"):FindChild("SellRightSide"):FindChild("SellSimilarItemsList")
		end
		for nKey, wndChild in ipairs(wndParent:GetChildren()) do
			local wnd = Apollo.LoadForm(self.xmlDoc, "WatchCommodityBtn", wndChild, self)
			wnd:SetAnchorOffsets(-59 + 36, 3 + 25, 0, 27 + 25)
		end
		--[[local wndHeader = tMarketPlaceCommodity.wndMain:FindChild("MetalHeader")
		Apollo.LoadForm(self.xmlDoc, "AddSearchButton", wndHeader, self)--]]
    end

	local fnOldOnSellListItemCheck = mc.OnSellListItemCheck
	mc.OnSellListItemCheck = function(tMarketplaceAuction, wndHandler, wndControl)
		fnOldOnSellListItemCheck(tMarketplaceAuction, wndHandler, wndControl)
		local wndParent = tMarketplaceAuction.wndMain:FindChild("SellRightSide")
		local wnd = Apollo.LoadForm(self.xmlDoc, "WatchCommodityBtn", wndParent, self)
		wnd:SetAnchorOffsets(-59 + 36, 3 + 25, 0, 27 + 25)
		wnd:SetData(wndHandler:GetData())
	end

	Apollo.RegisterSlashCommand("flipauction", "Open", self)
	Apollo.RegisterEventHandler("PostItemAuctionResult", "OnPostItemAuctionResult", self)
end

function FlipAuction:ToSave()
	return { 
		nFilterAmount = self.nFilterAmount,
	}
end

function FlipAuction:RestoreSave( tSave )
	tSave = tSave or {}
	self:SetFilterAmount(tSave.nFilterAmount or 0)
end

function FlipAuction:AddWindow( strName, wndParent, objHandler)
	return self.market:AddWindow(strName, wndParent, objHandler)
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipAuction:Open()
	self.groupAH = self.groups:GetAH()
	
	-- setup the tabs
	self.wndTabs:DestroyChildren()
	local tab1 = self:AddWindow("Tab", self.wndTabs, self)
	tab1:SetText("AH Overview")
	tab1:SetData(self.OnOpenOverviewTab)
	self.wndTabs:ArrangeChildrenHorz(0)

	self:OnTab(tab1)
	self.wndMain:Invoke()
end

function FlipAuction:OnClose()
	for nKey, updater in ipairs(self.tUpdaters) do 
		updater:Destroy()
	end
end	

-----------------------------------------------------------------------------------------------
-- tabbage

FlipAuction.Tabs = {
	OVERVIEW = 1,
}

function FlipAuction:OnTab( tab )
	for nKey, tab1 in ipairs(self.wndTabs:GetChildren()) do
		if tab1 ~= tab then tab1:SetOpacity(0.7) end
	end
	for nKey, updater in ipairs(self.tUpdaters) do 
		updater:Destroy()
	end
	self.tUpdaters = {}
	tab:SetOpacity(1)
	tab:GetData()(self)
	
	Apollo.StartTimer("AuctionTimerDelay")
end

function FlipAuction:OnOpenOverviewTab()
	self.tabActive = FlipAuction.Tabs.OVERVIEW
	self.headerActive = nil

	-- setup the sub tab
	self.wndSubTab:DestroyChildren()
	local wnd = Apollo.LoadForm(self.xmlDoc, "OverviewSubTab", self.wndSubTab, self)
	wnd:FindChild("RefreshContinuous"):Show(false)
	wnd:FindChild("CollapseAll"):Show(false)
	wnd:FindChild("ExpandAll"):Show(false)
	wnd:FindChild("CancelUndercut"):Show(false)
	wnd:FindChild("FilterContainer"):Show(false)
	if not self.market:SplitCancel() then wnd:FindChild("CancelUndercut"):SetTooltip("") end

	-- the filter window for squashing
	--self.wndFilter = self.wndSubTab:FindChild("Filtering")
	--self:SetFilterAmount(self.nFilterAmount)

	-- clear the left side
	self.wndLeft:DestroyChildren()
	
	-- draw the CX group header on the left side and act like we toggled it
	local header = self.groupAH:Draw(self.wndLeft, self)
	if self.wndLeft:FindChild("GroupHeaderButton") then
		self:OnGroupHeaderButtonDown(self.wndLeft:FindChild("GroupHeaderButton"))
	end
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipAuction:Dirty()
	Apollo.StopTimer("AuctionTimerDirty")
	Apollo.StartTimer("AuctionTimerDirty")
end

-- redraw the window; make sure we're still on the window
function FlipAuction:Redraw()
	local nPos = self.wndRight:GetVScrollPos()

	self.wndRight:ArrangeChildrenVert(0)
	local nRange = self.wndRight:GetVScrollRange()
	if nPos > nRange then 
		self.wndRight:SetVScrollPos(nRange)
	end
end

function FlipAuction:SortHeadersByKey()
	self.wndRight:ArrangeChildrenVert(0, function(wnd1, wnd2)
		local n1 = wnd1:GetData():GetSortKey()
		local n2 = wnd2:GetData():GetSortKey()
		return n1 < n2
	end)
end	

-- called by the + button in the CX
function FlipAuction:AddWatched(tIds)
	for nKey, nId in ipairs(tIds) do
		self.groups:AddItem(nId)
	end
	if self.headerActive then
		self:DisplayItems(self.headerActive:GetGroup())
	end
	if self.bCollapsed then
		self:OnCollapse()
	end
	-- self.market:RefreshAH()
end

function FlipAuction:OnAddCommodity( wndHandler, wndControl, eMouseButton )
	local itemData = nil
	if wndHandler:GetData() then
		itemData = wndHandler:GetData()
	else
		local aucCurr = wndHandler:GetParent():GetData()
		itemData = aucCurr:GetItem()
	end
	local nId = itemData:GetItemId()
	self:AddWatched({ nId })
end

-- called by the + button by the search
function FlipAuction:OnAddSearch( wndHandler, wndControl, eMouseButton )
	--[[local wndMain = wndHandler:GetParent():GetParent()
	local tChildren = wndMain:FindChild("MainScrollContainer"):GetChildren()
	local tIds = {}
	for _, wndChild in ipairs(tChildren) do
		-- if it's not the stupid "too many search results" window, then add a +
		if wndChild:GetText() == "" then
			local nId = tonumber(wndChild:GetName())
			table.insert(tIds, nId)
		end
	end
	self:AddCommodities(tIds)--]]
end

-- OnGroupHeaderButtonDown : HeaderButton -> 
-- 
function FlipAuction:OnGroupHeaderButtonDown( btn )
	local header = btn:GetParent():GetData()
	if header ~= self.headerActive then
		if self.headerActive then
			self.headerActive:Unfocus()
			self.headerActive:CollapseItems()
		end
		self.headerActive = header
		header:Focus()
		header:ExpandSubgroups()
		self:DisplayItems(header:GetGroup())
	end
end

function FlipAuction:DisplayItems( group )
	self.groupActive = group
	
	for nKey, updater in ipairs(self.tUpdaters) do 
		updater:Destroy()
	end
	self.tUpdaters = {}

	local tfns = {
		[FlipAuction.Tabs.OVERVIEW] = fnDrawItemOverview,
		--[FlipAuction.Tabs.SELLORDER] = fnDrawItemSellOrder,
	}

	local unitPlayer = GameLib.GetPlayerUnit()
	for idx, itemCurr in ipairs(unitPlayer:GetAuctionableItems()) do
		local item = group:GetItem(itemCurr:GetItemId())
		if item then
			local updater = FlipAuctionUpdater:new({
				item = item,
				itemData = itemCurr,
				auction = self,
				market = self.market,
				wndParent = self.wndRight,
				tab = self.tabActive,
			})
			table.insert(self.tUpdaters, updater)
		end
	end

	self:SortHeadersByKey()
	if self.bCollapsed then self:OnCollapse() end
end

-- OnCollapseAll : ->
-- collapse all the orders, go thru and toggle off everything
function FlipAuction:OnCollapse()
	self.bCollapsed = true
	for nKey, updater in ipairs(self.tUpdaters) do
		updater:CollapseListings()
	end
	self:Redraw()
end

function FlipAuction:OnExpand()
	self.bCollapsed = false
	for nKey, updater in ipairs(self.tUpdaters) do
		updater:ExpandListings()
	end
	self:Redraw()
end

-- OnRefresh : RefreshButton Window MouseClick ->
-- this is called when the "Refresh" button is clicked
-- asks for commodity info and commodity orders
function FlipAuction:OnRefresh( btnRefresh, wndControl, eMouseButton )
	self:DisplayItems(self.groupActive)
	self.market:RefreshAH()
end

-- GetFilterAmount : -> Number
-- get the value of the filter slider
function FlipAuction:GetFilterAmount()
	--return self.nFilterAmount or self.wndFilter and self.wndFilter:GetValue() or 0
end

-- SetFilterAmount : Number -> Number
-- set the value of the filter slider and return it
function FlipAuction:SetFilterAmount(nFilterAmount)
	--[[if self.wndFilter then
		self.wndFilter:SetValue(nFilterAmount)
	end
	self.nFilterAmount = nFilterAmount
	return nFilterAmount--]]
end

-- OnFilterChange : FilterWindow Window Float Float -> 
function FlipAuction:OnFilterChange( wndHandler, wndControl, fNewValue, fOldValue )
	--[[self.nFilterAmount = fNewValue
	if fNewValue ~= fOldValue then
		for nKey, updater in pairs(self.tUpdaters) do
			updater:Redraw(fNewValue)
		end
		self:SortHeadersByKey()
		if self.bCollapsed then self:OnCollapse() end
	end--]]
end

-- OnCancelAllUndercut : ->
-- cancel every undercut order
function FlipAuction:OnCancelUndercut( wndHandler, wndControl, eMouseButton)
	--[[local bIsBuy = (eMouseButton == 0)
	local bSplitCancel = self.market:SplitCancel()
    for nId, updater in pairs(self.tUpdaters) do
		updater:CancelAllUndercut(bIsBuy)
		if not bSplitCancel then updater:CancelAllUndercut(not bIsBuy) end
	end
	self:SortHeadersByKey()
	self:StartRefreshTimer()--]]
end

-- StartRefreshTimer : ->
-- starts the timer to delay refreshing
function FlipAuction:StartRefreshTimer()
	--Apollo.StartTimer("TimerDelay")
end

-- OnTimerDelay : ->
-- this is the handler for the small delay for refreshing stats after cancelling
function FlipAuction:OnTimerDelay() 
	--Apollo.StopTimer("TimerDelay")
	--self.market:RefreshCX()
end

function FlipAuction:OnPostItemAuctionResult()
	self:DisplayItems(self.groupActive)
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipAuctionUpdater:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
	return o
end

-- TODO: do we REALLY need all these fields jeez
function FlipAuctionUpdater:Init()
	self.item = self.item
	self.header = self.item:Draw(self.wndParent, self, self.itemData)
	self.market = self.market 
	self.auction = self.auction
	self.tab = self.tab

	self.nId = self.item:GetItemId()
	self.itemData = self.itemData or Item.GetDataFromId(self.nId)

	-- we have no orders
	self.tListings = {}

	-- tell the market to update this nId
	self.market:AddSub(self.nId, self)

	-- initialize the window
	self.wnd = self.header:GetWindow()
	--self.wnd:SetSprite("BK3:UI_BK3_ItemQualityWhite")

	local container = self.auction:AddWindow("Container", self.wnd, self)
	self.wnd:FindChild("ItemName"):SetAnchorPoints(0, 0, 0.37, 0)

	self.wndBid = self.auction:AddWindow("OrderContainer", container, self)
	self.wndBid:FindChild("Cancel"):Show(false)
	self.wndBid:FindChild("Confirm"):Show(false)
	self.wndBid:FindChild("Quantity"):Show(false)
	self.wndBuyout = self.auction:AddWindow("OrderContainer", container, self)
	self.wndBuyout:FindChild("Cancel"):Show(false)
	self.wndBuyout:FindChild("Quantity"):Show(false)
	container:ArrangeChildrenHorz(0)
	
	-- our listings window
	self.wndListings = self.wnd:FindChild("ItemMembers")
	
	-- the price associated with this commodity (probably nil at this point)	
	self.tPrice = self.tPrice or self.market:GetPrice(self.nId)
	self:Redraw()
end

function FlipAuctionUpdater:Destroy()
	self.wnd:Destroy()
	self.market:RemoveSub(self.nId, self)
end

function FlipAuctionUpdater:Notify( nId, tPrice )
	self:Redraw(self.auction:GetFilterAmount(), tPrice)
end

function FlipAuctionUpdater:OnAddCommodity()
	--self.commodity:AddCommodities({ self.nId })
end

function FlipAuctionUpdater:Redraw( nFilterAmount, tPrice )
	local price = tPrice or self.tPrice 
	if not price then return nil end
	
	if price.nBuyout then 
		self.wndBid:FindChild("Price"):SetAmount(math.max(price.nBuyout - 1, 0))
		self.wndBuyout:FindChild("Price"):SetAmount(price.nBuyout)
		local cFee = MarketplaceLib.GetItemAuctionCost(self.itemData)
		if cFee then price.nFee = cFee:GetAmount() end
	end

	--[[local options = self.item:GetOptions("MARKET")
	local strBuyOptionName, nBuyOptionPrice, nBuyQuantity = options:GetBuyOptions(price)
	local strSellOptionName, nSellOptionPrice, nSellQuantity = options:GetSellOptions(price)
	
	if nSellQuantity and nSellQuantity < 0 then
		nSellQuantity = math.abs(math.max(self.itemData:GetBackpackCount() + nSellQuantity, 0))
	else
		nSellQuantity = self.itemData:GetBackpackCount() or 0
	end

	if self.wndBuyContainer then
		self.wndBuyContainer:SetTooltip(strBuyOptionName)
		self.wndBuyContainer:FindChild("Price"):SetAmount(nBuyOptionPrice)
		self.wndBuyContainer:FindChild("Quantity"):SetText(nBuyQuantity)
	end

	self.wndSellContainer:SetTooltip(strSellOptionName)
	self.wndSellContainer:FindChild("Price"):SetAmount(nSellOptionPrice)
	self.wndSellContainer:FindChild("Quantity"):SetText(nSellQuantity)--]]

	self.header:SetSortKey(self.itemData:GetName())
	self.tPrice = price

	--[[self:Recolor(nFilterAmount, nBuyOptionPrice, nSellOptionPrice)--]]
end

function FlipAuctionUpdater:Recolor( nFilterAmount, nBuy, nSell )
	--[[local options = self.item:GetOptions("MARKET")
	local strColorName, bSquashed, strColor = options:GetColorOptions(nFilterAmount, nBuy, nSell)

	if self.tPrice then
		local nSellVendor = self.tPrice.nSellVendor or 0
		if self.wndBuyContainer then
			self.wndBuyContainer:FindChild("Price"):SetTextColor(
				nBuy < nSellVendor and "red" or "white")
		end
		self.wndSellContainer:FindChild("Price"):SetTextColor(
			(nSell * 0.88) < nSellVendor and "red" or "white")
	end

	-- set the border color
	if self.tab == FlipAuction.Tabs.OVERVIEW then
		self.header:SetBorderColor(strColor)
	
		-- update squashing
		self.header:ToggleSquash(bSquashed)
	end--]]
end

function FlipAuctionUpdater:OnRequestStatistics( btn )
	--self.commodity:RequestStatistics(self.nId)
end

-- OnConfirmMouseEnter : ConfirmButton ->
function FlipAuctionUpdater:OnMouseEnterConfirm( btnConfirm )
	--local bIsBuy = (btnConfirm:GetParent() == self.wndBuyContainer)
	self:AttemptEnableTransaction(btnConfirm)
end

-- AttemptEnableTransaction : Boolean ConfirmButton -> 
-- attempt to enable the transaction on mouseover
function FlipAuctionUpdater:AttemptEnableTransaction( btnConfirm )
	-- the amount we're ordering
	-- local nQuantity = tonumber(container:FindChild("Quantity"):GetText()) or 0

	-- the price (let's make sure we use the same one being displayed)
	local cBid = self.wndBid:FindChild("Price"):GetCurrency() 
	local cBuyout = self.wndBuyout:FindChild("Price"):GetCurrency()

	if cBuyout:GetAmount() > 0 then
		-- find the execute button, we're going to turn it on
		btnConfirm:Enable(true)
		local eType = GameLib.CodeEnumConfirmButtonType.MarketplaceAuctionSellSubmit
		btnConfirm:SetActionData(eType, self.itemData, cBid, cBuyout)
	end 
end

function FlipAuctionUpdater:OnOrderPriceChange( wndHandler )
	--local nBuy = self.wndBuyContainer:FindChild("Price"):GetAmount()
	--local nSell = self.wndSellContainer:FindChild("Price"):GetAmount()
	--wndHandler:GetParent():GetParent():SetTooltip("Price overriden")
	--self:Recolor(0, nBuy, nSell)
end

function FlipAuctionUpdater:AddListing( tCurrOrder, tListings )
	--[[local listingNew
	for nKey, listing in ipairs(tListings) do
		if listing:Update(tCurrOrder) then listingNew = listing break
		end
	end
	listingNew = listingNew or FlipAuctionListing:new({ 
		tCurrOrder = tCurrOrder, 
		updater = self,
		market = self.market,
		header = self.header,
		commodity = self.commodity,
		tab = self.tab,
	})
	table.insert(self.tListings, listingNew)
	self.header:ToggleMembers()
	self.header:ToggleMembers()--]]
end

function FlipAuctionUpdater:ClearListing( listing1 )
	--[[for nKey, listing2 in ipairs(self.tListings) do
		if listing1 == listing2 then
			table.remove(self.tListings, nKey)
			return nil
		end
	end--]]
end

function FlipAuctionUpdater:ClearListings()
	--[[local tListings = self.tListings
	for nKey, listing in ipairs(self.tListings) do
		listing:Clear()
	end
	self.tListings = {}
	return tListings--]]
end

function FlipAuctionUpdater:CancelAllUndercut( bIsBuy )
	--[[local tUndercut = {}
	for nKey, listing in ipairs(self.tListings) do
		if listing:IsBuy() == bIsBuy and listing:IsUndercut() then 
			table.insert(tUndercut, listing) 
		end
	end
	for nKey, listing in ipairs(tUndercut) do
		listing:Cancel()
	end
	self:ToggleListings()
	self:ToggleListings()--]]
end

-- GetPrice : -> PriceTable
-- the current price we're using to display
function FlipAuctionUpdater:GetPrice() 
	--return self.tPrice
end

function FlipAuctionUpdater:ToggleListings()
	--self.header:ToggleMembers()
end

function FlipAuctionUpdater:ExpandListings()
	--self.header:ExpandMembers()
end

function FlipAuctionUpdater:CollapseListings()
	--self.header:CollapseMembers()
end

function FlipAuctionUpdater:OnGroupItemButton( btn, wndControl, eMouseButton )
	--[[if eMouseButton ~= 1 then
		self.header:ToggleMembers()
		self.commodity:Redraw()
	else
		if self.market:OneClickDelete() then
			self.item:Destroy()
			self:Destroy()
			self.commodity:Redraw()
		else
			self.item:PullUpOptions()
		end
	end--]]
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

Apollo.RegisterPackage(FlipAuction, "FlipAuction", 1, {})
