-----------------------------------------------------------------------------------------------
--[[ Copyright (c) chckxy (watdametz@gmail.com). All rights reserved

This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.  --]]
-----------------------------------------------------------------------------------------------

local FlipCommodity = {}
local FlipCommodityUpdater = {}
local FlipCommodityListing = {}

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipCommodity:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
	return o
end

function FlipCommodity:Init()
	-- we have a market
	self.market = self.market
	
	-- we have the groups module
	self.groups = self.groups

	-- headers being displayed
	self.tUpdaters = {}
	
	-- we have the main window but we don't know if they're going to open the CX
	self.xmlDoc = XmlDoc.CreateFromFile("Flip.xml")	
	self.wndMain = self.wndMain
	self.wndTabs = self.wndMain:FindChild("TabContainer")
	self.wndSubTab = self.wndMain:FindChild("SubTabContainer")
	self.wndLeft = self.wndMain:FindChild("Left")
	self.wndRight = self.wndMain:FindChild("Right")

	-- might have commodity stats installed
	-- depends on RequestStatistics not needing the cstat instance
	local cstats = Apollo.GetAddon("CommodityStats")
	self.RequestStatistics = function(tMarket, nId) cstats.RequestStatistics(nil, nId) end

	-- timers for (refresh) delay
	self.timerDelay = Apollo.CreateTimer("TimerDelay", 2, false)
	Apollo.StopTimer("TimerDelay")
	self.timerRefresh = Apollo.CreateTimer("TimerRefresh", 30, true)
	Apollo.StopTimer("TimerRefresh")
	Apollo.RegisterTimerHandler("TimerDelay", "OnTimerDelay", self)
	Apollo.RegisterTimerHandler("TimerRefresh", "RefreshCX", self.market)

	-- track our orders
	Apollo.RegisterEventHandler("OwnedCommodityOrders", "OnOwnedCommodityOrders", self)
	
	-- our filter is at zero initially
	self.nFilterAmount = 0
	
	-- add the watch button to all the commodities; got idea from CommodityStats
	self.bHideDefaultPostMessage = true
	local mc = Apollo.GetAddon("MarketplaceCommodity")
	local fnOldHeaderBtnToggle = mc.OnHeaderBtnToggle
	mc.OnHeaderBtnToggle = function(tMarketPlaceCommodity)
		fnOldHeaderBtnToggle(tMarketPlaceCommodity)
		local tChildren = tMarketPlaceCommodity.wndMain:FindChild("MainScrollContainer"):GetChildren()
		for _, wndChild in ipairs(tChildren) do
			if wndChild:GetText() == "" then
				Apollo.LoadForm(self.xmlDoc, "WatchCommodityBtn", wndChild, self)
			end
		end
		local wndHeader = tMarketPlaceCommodity.wndMain:FindChild("MetalHeader")
		Apollo.LoadForm(self.xmlDoc, "AddSearchButton", wndHeader, self)
    end

	if cstats then
		self.fnOldOnPostCustomMessage = cstats.OnPostCustomMessage
		self.fnNewOnPostCustomMessage = function(cstats, mc, strMessage, bResultOK, nDuration)
			Print(strMessage)
		end
		cstats.OnPostCustomMessage = self.fnNewOnPostCustomMessage
	else
		self.fnOldOnPostCustomMessage = mc.OnPostCustomMessage
		self.fnNewOnPostCustomMessage = function(mc, strMessage, bResultOK, nDuration)
			Print(strMessage)
		end
		mc.OnPostCustomMessage = self.fnNewOnPostCustomMessage
	end

	-- are we collapsed/expanded
	self.bCollapsed = false
	self.cstats = cstats and true
end

function FlipCommodity:ToSave()
	return { 
		bCollapsed = self.bCollapsed,
		nFilterAmount = self.nFilterAmount,
	}
end

function FlipCommodity:RestoreSave( tSave )
	tSave = tSave or {}
	self.bCollapsed = tSave.bCollapsed or false
	self:SetFilterAmount(tSave.nFilterAmount or 0)
end

function FlipCommodity:AddWindow( strName, wndParent, objHandler)
	return self.market:AddWindow(strName, wndParent, objHandler)
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------
-- dealing with the CX side of the /flip window

-- called by /flip
function FlipCommodity:Open()
	self.groupCX = self.groups:GetCX()
	
	-- setup the tabs
	self.wndTabs:DestroyChildren()
	local tab1 = self:AddWindow("Tab", self.wndTabs, self)
	tab1:SetText("CX Overview")
	tab1:SetData(self.OnOpenOverviewTab)
	local tab2 = self:AddWindow("Tab", self.wndTabs, self)
	tab2:SetText("Place Sell Order")
	tab2:SetData(self.OnOpenSellOrderTab)
	if self.cstats then
		local tab3 = self:AddWindow("Tab", self.wndTabs, self)
		tab3:SetText("Scan For Profit")
		tab3:SetData(self.OnOpenScanForProfit)
	end
		self.wndTabs:ArrangeChildrenHorz(0)

	self:OnTab(tab1)
	self.wndMain:Invoke()
end

function FlipCommodity:OnClose()
	for nId, updater in pairs(self.tUpdaters) do 
		updater:Destroy()
	end
end	

-----------------------------------------------------------------------------------------------
-- tabbage

FlipCommodity.Tabs = {
	OVERVIEW = 1,
	SELLORDER = 2,
	SCANPROFIT = 3,
}

function FlipCommodity:OnTab( tab )
	for nKey, tab1 in ipairs(self.wndTabs:GetChildren()) do
		if tab1 ~= tab then tab1:SetOpacity(0.7) end
	end
	for nId, updater in pairs(self.tUpdaters) do 
		updater:Destroy()
	end
	self.tUpdaters = {}
	tab:SetOpacity(1)
	tab:GetData()(self)
end

function FlipCommodity:OnOpenOverviewTab()
	self.tabActive = FlipCommodity.Tabs.OVERVIEW
	self.headerActive = nil

	-- setup the sub tab
	self.wndSubTab:DestroyChildren()
	local wnd = Apollo.LoadForm(self.xmlDoc, "OverviewSubTab", self.wndSubTab, self)
	wnd:FindChild("RefreshContinuous"):Show(false)
	if not self.market:SplitCancel() then
		wnd:FindChild("CancelUndercut"):SetTooltip("")
	end

	-- the filter window for squashing
	self.wndFilter = self.wndSubTab:FindChild("Filtering")
	self:SetFilterAmount(self.nFilterAmount)

	-- clear the left side
	self.wndLeft:DestroyChildren()
	
	-- draw the CX group header on the left side and act like we toggled it
	local header = self.groupCX:Draw(self.wndLeft, self)
	if self.wndLeft:FindChild("GroupHeaderButton") then
		self:OnGroupHeaderButtonDown(self.wndLeft:FindChild("GroupHeaderButton"))
	end
end

function FlipCommodity:OnOpenSellOrderTab()
	self.tabActive = FlipCommodity.Tabs.SELLORDER
	self.headerActive = nil

	-- setup the sub tab
	self.wndSubTab:DestroyChildren()
	local wnd = Apollo.LoadForm(self.xmlDoc, "SellOrderSubTab", self.wndSubTab, self)
	wnd:FindChild("RefreshContinuous"):Show(false)

	self.wndLeft:DestroyChildren()

	-- draw the CX group header on the left side and act like we toggled it
	local header = self.groupCX:Draw(self.wndLeft, self)
	if self.wndLeft:FindChild("GroupHeaderButton") then
		self:OnGroupHeaderButtonDown(self.wndLeft:FindChild("GroupHeaderButton"))
	end
end

function FlipCommodity:OnOpenScanForProfit()
	self.tabActive = FlipCommodity.Tabs.SCANPROFIT
	self.headerActive = nil
	self.wndLeft:DestroyChildren()
	self.wndSubTab:DestroyChildren()

	local cstats = Apollo.GetAddon("CommodityStats")
	if cstats.statistics then self:DisplayProfitable(cstats.statistics) end
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

-- redraw the window; make sure we're still on the window
function FlipCommodity:Redraw()
	local nPos = self.wndRight:GetVScrollPos()

	self.wndRight:ArrangeChildrenVert(0)
	local nRange = self.wndRight:GetVScrollRange()
	if nPos > nRange then 
		self.wndRight:SetVScrollPos(nRange)
	end
end

function FlipCommodity:SortHeadersByKey()
	self.wndRight:ArrangeChildrenVert(0, function(wnd1, wnd2)
		local n1 = wnd1:GetData():GetSortKey()
		local n2 = wnd2:GetData():GetSortKey()
		return n1 > n2
	end)
end	

-- called by the + button in the CX
function FlipCommodity:AddCommodities(tIds)
	for nKey, nId in ipairs(tIds) do
		self.groups:AddItem(nId)
	end
	if self.headerActive then
		self:DisplayItems(self.headerActive:GetGroup())
	end
	if self.bCollapsed then
		self:OnCollapse()
	end
	self.market:RefreshCX()
end

function FlipCommodity:OnAddCommodity( wndHandler, wndControl, eMouseButton )
	local nId = tonumber(wndControl:GetParent():GetName())
	self:AddCommodities({ nId })
end

-- called by the + button by the search
function FlipCommodity:OnAddSearch( wndHandler, wndControl, eMouseButton )
	local wndMain = wndHandler:GetParent():GetParent()
	local tChildren = wndMain:FindChild("MainScrollContainer"):GetChildren()
	local tIds = {}
	for _, wndChild in ipairs(tChildren) do
		-- if it's not the stupid "too many search results" window, then add a +
		if wndChild:GetText() == "" then
			local nId = tonumber(wndChild:GetName())
			table.insert(tIds, nId)
		end
	end
	self:AddCommodities(tIds)
end

-- OnGroupHeaderButtonDown : HeaderButton -> 
-- 
function FlipCommodity:OnGroupHeaderButtonDown( btn )
	local header = btn:GetParent():GetData()
	if header ~= self.headerActive then
		if self.headerActive then
			self.headerActive:Unfocus()
			self.headerActive:CollapseItems()
		end
		self.headerActive = header
		header:Focus()
		header:ExpandSubgroups()
		self:DisplayItems(header:GetGroup())
		self.market:RefreshCX()
	end
end 

local fnDrawItemOverview = function(item)
	return true
end

local fnDrawItemSellOrder = function(item)
	return item:GetData():GetBackpackCount() > 0
end

function FlipCommodity:DisplayItems( group )
	self.groupActive = group
	
	for nId, updater in pairs(self.tUpdaters) do 
		updater:Destroy()
	end
	self.tUpdaters = {}

	local tfns = {
		[FlipCommodity.Tabs.OVERVIEW] = fnDrawItemOverview,
		[FlipCommodity.Tabs.SELLORDER] = fnDrawItemSellOrder,
	}

	local fnDrawItem = function(item)
		if tfns[self.tabActive](item) then
			local updater = FlipCommodityUpdater:new({
				item = item,
				commodity = self,
				market = self.market,
				bLoadCStats = self.cstats,
				wndParent = self.wndRight,
				tab = self.tabActive,
			})
			table.insert(self.tUpdaters, item:GetItemId(), updater)
		end
	end

	local fnDrawAllItems
	fnDrawAllItems = function(subgroup)
		subgroup:ForEachItem(fnDrawItem)
		subgroup:ForEachSubgroup(fnDrawAllItems)
	end

	fnDrawAllItems(group)
	self:SortHeadersByKey()
	if self.bCollapsed then self:OnCollapse() end
end

function FlipCommodity:DisplayProfitable( tStats )
	for nId, stats in pairs(tStats) do
		self:CheckProfitable(nId, stats)
	end
	self:SortHeadersByKey()
end

function FlipCommodity:CheckProfitable(nId, stats)
	local lastTime = table.maxn(stats)
	if lastTime == 0 then return nil end
	if stats[lastTime].sellOrderCount == 0 then return nil end
	local itemData = Item.GetDataFromId(nId)
	if itemData == nil then return nil end
	
	local tStats = stats[lastTime]
	local nBuyOrders = tStats.buyOrderCount or 0
	local nSellOrders = tStats.sellOrderCount or 0

	local nOrders = math.min(nBuyOrders, nSellOrders)
	local nBuy = 0
	local nSell = 0
	if nBuyOrders >= 50 and nSellOrders >= 50 then
		nBuy = tStats.buyPrices.top50
		nSell = tStats.sellPrices.top50
	elseif nBuyOrders >= 10 and nSellOrders >= 10 then
		nBuy = tStats.buyPrices.top10
		nSell = tStats.sellPrices.top10
	elseif nBuyOrders >= 1 and nSellOrders >= 1 then
		nBuy = tStats.buyPrices.top1
		nSell = tStats.sellPrices.top1
	end
	
	-- % price difference, bigger is better
	local nDiff = (nSell * 0.88) - (nBuy * 1.02)
	local nPercentDiffPrice = nDiff / ((nBuy + nSell) / 2)

	local cVendor = itemData:GetSellPrice()
	local nVendor = (cVendor and cVendor:GetAmount()) or 0
	
	-- % quantity difference, smaller is better
	local nQuantityAvg = ((nSellOrders + nBuyOrders) / 2)
	local nQuantityDiff = math.abs(nSellOrders - nBuyOrders)
	local nPercentDiffQuantity = (nQuantityDiff / nQuantityAvg)

	local nScore = nPercentDiffPrice / nPercentDiffQuantity * (nOrders / 200)
	if (nScore > 0) and
	   (nDiff > 500) and 
	   (tStats.buyPrices.top1 and (tStats.buyPrices.top1 > nVendor)) and
	   (nQuantityAvg > 5) then 
		local updater = FlipCommodityUpdater:new({
			item = self.groups:MakeItem(nId),
			commodity = self,
			market = self.market,
			bLoadCStats = self.cstats,
			wndParent = self.wndRight,
			tab = self.tabActive,
			tStats = stats[lastTime],
			nKey = nScore,
		})
		table.insert(self.tUpdaters, nId, updater)
	end
end

-- this is called when we get updated information about our commodity orders;
-- delete all the old orders (since I know of no way to directly update them)
-- and then add all the new orders
function FlipCommodity:OnOwnedCommodityOrders( tOrders )
	local tListings = {}
 	for nKey, tCurrOrder in pairs(tOrders) do  -- key is the listing# not the item#
 		local nId = tCurrOrder:GetItem():GetItemId()
 		local updater = self.tUpdaters[nId]
 		if self.tabActive == FlipCommodity.Tabs.OVERVIEW or 
 		   self.tabActive == FlipCommodity.Tabs.SELLORDER and not tCurrOrder:IsBuy() then
			if updater then
				tListings[nId] = tListings[nId] or updater:ClearListings()
				updater:AddListing(tCurrOrder, tListings[nId])
			end
 		end
 	end

 	self:SortHeadersByKey()
 	if self.bCollapsed then self:OnCollapse() end
end

-- OnCollapseAll : ->
-- collapse all the orders, go thru and toggle off everything
function FlipCommodity:OnCollapse()
	self.bCollapsed = true
	for nId, updater in pairs(self.tUpdaters) do
		updater:CollapseListings()
	end
	self:Redraw()
end

function FlipCommodity:OnExpand()
	self.bCollapsed = false
	for nId, updater in pairs(self.tUpdaters) do
		updater:ExpandListings()
	end
	self:Redraw()
end

-- OnRefresh : RefreshButton Window MouseClick ->
-- this is called when the "Refresh" button is clicked
-- asks for commodity info and commodity orders
function FlipCommodity:OnRefresh( btnRefresh, wndControl, eMouseButton )
	self.market:RefreshCX()
	if self.bRefreshContinuous then
		self.bRefreshContinuous = false
		btnRefresh:SetText("Refresh")
		btnRefresh:FindChild("RefreshContinuous"):Show(false)
		Apollo.StopTimer("TimerRefresh")
	elseif eMouseButton == 1 then 
		self.bRefreshContinuous = true
		btnRefresh:SetText("Refreshing Continually")
		btnRefresh:FindChild("RefreshContinuous"):Show(true)
		Apollo.StartTimer("TimerRefresh")
	end
end

-- GetFilterAmount : -> Number
-- get the value of the filter slider
function FlipCommodity:GetFilterAmount()
	return self.nFilterAmount or self.wndFilter and self.wndFilter:GetValue() or 0
end

-- SetFilterAmount : Number -> Number
-- set the value of the filter slider and return it
function FlipCommodity:SetFilterAmount(nFilterAmount)
	if self.wndFilter then
		self.wndFilter:SetValue(nFilterAmount)
	end
	self.nFilterAmount = nFilterAmount
	return nFilterAmount
end

-- OnFilterChange : FilterWindow Window Float Float -> 
function FlipCommodity:OnFilterChange( wndHandler, wndControl, fNewValue, fOldValue )
	self.nFilterAmount = fNewValue
	if fNewValue ~= fOldValue then
		for nKey, updater in pairs(self.tUpdaters) do
			updater:Redraw(fNewValue)
		end
		self:SortHeadersByKey()
		if self.bCollapsed then self:OnCollapse() end
	end
end

-- OnCancelAllUndercut : ->
-- cancel every undercut order
function FlipCommodity:OnCancelUndercut( wndHandler, wndControl, eMouseButton)
	local bIsBuy = (eMouseButton == 0)
	local bSplitCancel = self.market:SplitCancel()
    for nId, updater in pairs(self.tUpdaters) do
		updater:CancelAllUndercut(bIsBuy)
		if not bSplitCancel then updater:CancelAllUndercut(not bIsBuy) end
	end
	self:SortHeadersByKey()
	self:StartRefreshTimer()
end

-- StartRefreshTimer : ->
-- starts the timer to delay refreshing
function FlipCommodity:StartRefreshTimer()
	Apollo.StartTimer("TimerDelay")
end

-- OnTimerDelay : ->
-- this is the handler for the small delay for refreshing stats after cancelling
function FlipCommodity:OnTimerDelay() 
	Apollo.StopTimer("TimerDelay")
	self.market:RefreshCX()
end

function FlipCommodity:HidePostMessageToggle( bHide )
	if not bHide ~= self.bHideDefaultPostMessage then return end
	local obj = Apollo.GetAddon(self.cstats and "CommodityStats" or "MarketplaceCommodity")
	obj.OnPostCustomMessage = bHide and self.fnNewOnPostCustomMessage or self.fnOldOnPostCustomMessage
	self.bHideDefaultPostMessage = bHide
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipCommodityUpdater:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
	return o
end

-- TODO: do we REALLY need all these fields jeez
function FlipCommodityUpdater:Init()
	self.item = self.item
	self.header = self.item:Draw(self.wndParent, self)
	self.market = self.market 
	self.commodity = self.commodity
	self.tab = self.tab

	-- we also have an item id
	self.nId = self.item:GetItemId()

	-- we have no orders
	self.tListings = {}

	-- tell the market to update this nId
	if self.tab ~= FlipCommodity.Tabs.SCANPROFIT then
		self.market:AddSub(self.nId, self)
	end

	-- get the item data
	self.itemData = Item.GetDataFromId(self.nId)

	-- initialize the window
	self.wnd = self.header:GetWindow()
	self.wnd:SetSprite("BK3:UI_BK3_ItemQualityWhite")
	if self.tab ~= FlipCommodity.Tabs.OVERVIEW then
		self.wnd:SetBGColor("xkcdBluishGreen")
	end
	local container = self.commodity:AddWindow("Container", self.wnd, self)
	self.wnd:FindChild("ItemName"):SetAnchorPoints(0, 0, 0.37, 0)

	if self.tab ~= FlipCommodity.Tabs.SELLORDER then
		self.wndBuyContainer = self.commodity:AddWindow("OrderContainer", container, self)
		self.wndBuyContainer:FindChild("Cancel"):Show(false)
		if self.tab == FlipCommodity.Tabs.SCANPROFIT then
			self.wndBuyContainer:FindChild("Confirm"):Show(false)
		end
	end
	self.wndSellContainer = self.commodity:AddWindow("OrderContainer", container, self)
	self.wndSellContainer:FindChild("Cancel"):Show(false)
	if self.tab == FlipCommodity.Tabs.SCANPROFIT then
		self.wndSellContainer:FindChild("Confirm"):Show(false)
	end
	container:ArrangeChildrenHorz(0)

	if self.bLoadCStats then 
		self.commodity:AddWindow("StatButton", self.wnd:FindChild("GroupItemButton"), self) 
	end
	if self.tab == FlipCommodity.Tabs.SCANPROFIT then
		if not self.item:IsInGroup() then
			local wndAdd = self.commodity:AddWindow("AddProfitableBtn", container, self)
		end
	end

	-- our listings window
	self.wndListings = self.wnd:FindChild("ItemMembers")
	
	-- the price associated with this commodity (probably nil at this point)	
	self.tPrice = self.tPrice or self.market:GetPrice(self.nId)
	self:Redraw(self.commodity:GetFilterAmount())
end

function FlipCommodityUpdater:Destroy()
	self.wnd:Destroy()
	self.market:RemoveSub(self.nId, self)
end

function FlipCommodityUpdater:Notify( nId, tPrice )
	self:Redraw(self.commodity:GetFilterAmount(), tPrice)
end

function FlipCommodityUpdater:OnAddCommodity()
	self.commodity:AddCommodities({ self.nId })
end

function FlipCommodityUpdater:Redraw( nFilterAmount, tPrice )
	local price = tPrice or self.tPrice 

	if self.tab ~= FlipCommodity.Tabs.SCANPROFIT then
		if not price then return nil end
		local options = self.item:GetOptions("MARKET")
		local strBuyOptionName, nBuyOptionPrice, nBuyQuantity = options:GetBuyOptions(price)
		local strSellOptionName, nSellOptionPrice, nSellQuantity = options:GetSellOptions(price)
		
		if nSellQuantity and nSellQuantity < 0 then
			nSellQuantity = math.abs(math.max(self.itemData:GetBackpackCount() + nSellQuantity, 0))
		else
			nSellQuantity = self.itemData:GetBackpackCount() or 0
		end
	
		if self.wndBuyContainer then
			self.wndBuyContainer:SetTooltip(strBuyOptionName)
			self.wndBuyContainer:FindChild("Price"):SetAmount(nBuyOptionPrice)
			self.wndBuyContainer:FindChild("Quantity"):SetText(nBuyQuantity)
		end
	
		self.wndSellContainer:SetTooltip(strSellOptionName)
		self.wndSellContainer:FindChild("Price"):SetAmount(nSellOptionPrice)
		self.wndSellContainer:FindChild("Quantity"):SetText(nSellQuantity)
	
		if self.tab == FlipCommodity.Tabs.OVERVIEW then
			self.header:SetSortKey(((nSellOptionPrice * 0.88) - (nBuyOptionPrice * 1.02)) * nBuyQuantity)
		elseif self.tab == FlipCommodity.Tabs.SELLORDER then
			self.header:SetSortKey(nSellOptionPrice * (nSellQuantity))
		end
		self.tPrice = price
	
		self:Recolor(nFilterAmount, nBuyOptionPrice, nSellOptionPrice)
	else
		local nBuyOrders = self.tStats.buyOrderCount or 0
		local nSellOrders = self.tStats.sellOrderCount or 0
		self.wndBuyContainer:FindChild("Price"):SetAmount(self.tStats.buyPrices.top1)
		self.wndSellContainer:FindChild("Price"):SetAmount(self.tStats.sellPrices.top1)
		self.wndBuyContainer:FindChild("Quantity"):SetText(tostring(nBuyOrders))
		self.wndSellContainer:FindChild("Quantity"):SetText(tostring(nSellOrders))
		self.header:SetSortKey(self.nKey)
	end
end

function FlipCommodityUpdater:Recolor( nFilterAmount, nBuy, nSell )
	local options = self.item:GetOptions("MARKET")
	local strColorName, bSquashed, strColor = options:GetColorOptions(nFilterAmount, nBuy, nSell)

	if self.tPrice then
		local nSellVendor = self.tPrice.nSellVendor or 0
		if self.wndBuyContainer then
			self.wndBuyContainer:FindChild("Price"):SetTextColor(
				nBuy < nSellVendor and "red" or "white")
		end
		self.wndSellContainer:FindChild("Price"):SetTextColor(
			(nSell * 0.88) < nSellVendor and "red" or "white")
	end

	-- set the border color
	if self.tab == FlipCommodity.Tabs.OVERVIEW then
		self.header:SetBorderColor(strColor)
	
		-- update squashing
		self.header:ToggleSquash(bSquashed)
	end
end

function FlipCommodityUpdater:OnRequestStatistics( btn )
	self.commodity:RequestStatistics(self.nId)
end

-- OnConfirmMouseEnter : ConfirmButton ->
function FlipCommodityUpdater:OnMouseEnterConfirm( btnConfirm )
	local bIsBuy = (btnConfirm:GetParent() == self.wndBuyContainer)
	self:AttemptEnableTransaction(bIsBuy, btnConfirm)
end

-- AttemptEnableTransaction : Boolean ConfirmButton -> 
-- attempt to enable the transaction on mouseover
function FlipCommodityUpdater:AttemptEnableTransaction( bIsBuy, btnConfirm )
	local container = bIsBuy and self.wndBuyContainer or self.wndSellContainer

	-- the amount we're ordering
	local nQuantity = tonumber(container:FindChild("Quantity"):GetText()) or 0

	-- the price (let's make sure we use the same one being displayed)
	local cPricePerUnit = container:FindChild("Price"):GetCurrency() -- not an integer

	local order = bIsBuy and CommodityOrder.newBuyOrder(self.nId) or CommodityOrder.newSellOrder(self.nId)
	if nQuantity > 0 and cPricePerUnit:GetAmount() > 0 then
		-- the order we'll make when we press the button
		order:SetForceImmediate(false)
		order:SetCount(math.min(nQuantity, 200))
		order:SetPrices(cPricePerUnit)

		-- incantation to turn this into a real button (copied from MarketplaceCommodities)
		if order:CanPost() then
			-- find the execute button, we're going to turn it on
			btnConfirm:Enable(true)
			btnConfirm:SetActionData(GameLib.CodeEnumConfirmButtonType.MarketplaceCommoditiesSubmit, order)
		end
	end
end

function FlipCommodityUpdater:OnOrderPriceChange( wndHandler )
	local nBuy = self.wndBuyContainer:FindChild("Price"):GetAmount()
	local nSell = self.wndSellContainer:FindChild("Price"):GetAmount()
	wndHandler:GetParent():GetParent():SetTooltip("Price overriden")
	self:Recolor(0, nBuy, nSell)
end

function FlipCommodityUpdater:AddListing( tCurrOrder, tListings )
	local listingNew
	for nKey, listing in ipairs(tListings) do
		if listing:Update(tCurrOrder) then listingNew = listing break
		end
	end
	listingNew = listingNew or FlipCommodityListing:new({ 
		tCurrOrder = tCurrOrder, 
		updater = self,
		market = self.market,
		header = self.header,
		commodity = self.commodity,
		tab = self.tab,
	})
	table.insert(self.tListings, listingNew)
	self.header:ToggleMembers()
	self.header:ToggleMembers()
end

function FlipCommodityUpdater:ClearListing( listing1 )
	for nKey, listing2 in ipairs(self.tListings) do
		if listing1 == listing2 then
			table.remove(self.tListings, nKey)
			return nil
		end
	end
end

function FlipCommodityUpdater:ClearListings()
	local tListings = self.tListings
	for nKey, listing in ipairs(self.tListings) do
		listing:Clear()
	end
	self.tListings = {}
	return tListings
end

function FlipCommodityUpdater:CancelAllUndercut( bIsBuy )
	local tUndercut = {}
	for nKey, listing in ipairs(self.tListings) do
		if listing:IsBuy() == bIsBuy and listing:IsUndercut() then 
			table.insert(tUndercut, listing) 
		end
	end
	for nKey, listing in ipairs(tUndercut) do
		listing:Cancel()
	end
	self:ToggleListings()
	self:ToggleListings()
end

-- GetPrice : -> PriceTable
-- the current price we're using to display
function FlipCommodityUpdater:GetPrice() 
	return self.tPrice
end

function FlipCommodityUpdater:ToggleListings()
	self.header:ToggleMembers()
end

function FlipCommodityUpdater:ExpandListings()
	self.header:ExpandMembers()
end

function FlipCommodityUpdater:CollapseListings()
	self.header:CollapseMembers()
end

function FlipCommodityUpdater:OnGroupItemButton( btn, wndControl, eMouseButton )
	if eMouseButton ~= 1 then
		self.header:ToggleMembers()
		self.commodity:Redraw()
	else
		if self.market:OneClickDelete() then
			self.item:Destroy()
			self:Destroy()
			self.commodity:Redraw()
		else
			self.item:PullUpOptions()
		end
	end
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------
-- commodity listings

function FlipCommodityListing:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
	return o
end

function FlipCommodityListing:Init()
	
	self.tCurrOrder = self.tCurrOrder
	self.updater = self.updater
	self.header = self.header
	self.market = self.market
	self.commodity = self.commodity
	self.tab = self.tab
	
	-- our listing
	self.wnd = self.header:AddMember(self)
	self.bIsBuy = self.tCurrOrder:IsBuy()
	
	local container = self.market:AddWindow("Container", self.wnd, self)
	if self.tab == FlipCommodity.Tabs.OVERVIEW then
		self.wndBuyContainer = self.market:AddWindow("OrderContainer", container, self)
	end
	self.wndSellContainer = self.market:AddWindow("OrderContainer", container, self)
	container:ArrangeChildrenHorz(0)
	
	local wndContainer = (self.bIsBuy and self.wndBuyContainer) or self.wndSellContainer
	local wndNotContainer = (self.bIsBuy and self.wndSellContainer) or self.wndBuyContainer
	
	if wndNotContainer then wndNotContainer:Show(false) end
	
	-- don't display the trade confirm buttons
	wndContainer:FindChild("Confirm"):Show(false)
	wndContainer:Show(true)
	wndContainer:FindChild("Cancel"):Show(true)
	
	-- update the list item
	local nCount = self.tCurrOrder:GetCount()
	local nPrice = self.tCurrOrder:GetPricePerUnit():GetAmount()

	wndContainer:FindChild("Price"):SetAmount(nPrice)
	wndContainer:FindChild("Quantity"):SetText(nCount)
	local tPrice = self.updater:GetPrice()
	-- TODO: this will break with price options
	local nCompetition = self.bIsBuy and tPrice.nBuy1Price or tPrice.nSell1Price
	self.bUndercut = (
		(self.bIsBuy       and nPrice < nCompetition) or 
		((not self.bIsBuy) and nPrice > nCompetition)
	)
	if self.bUndercut then wndContainer:FindChild("Price"):SetTextColor("red") end

	-- we don't need to show the item name
	self.wnd:FindChild("ItemName"):SetText("")
end

-- OnCancel : CancelButton ->
function FlipCommodityListing:OnCancel()
	self:Cancel()
	self.updater:ToggleListings()
	self.updater:ToggleListings()
	self.commodity:StartRefreshTimer()
end

function FlipCommodityListing:Cancel()
	self.tCurrOrder:Cancel()
	self.wnd:Destroy()
	self.updater:ClearListing(self)
end

function FlipCommodityListing:Destroy()
	self.wnd:Destroy()
end

function FlipCommodityListing:IsBuy()
	return self.bIsBuy
end

function FlipCommodityListing:IsUndercut()
	return self.bUndercut
end

function FlipCommodityListing:Clear()
	self.wnd:Destroy()
end

function FlipCommodityListing:Update( tCurrOrder )
	if self.tCurrOrder == tCurrOrder then self:Init() return true end
	return false
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

Apollo.RegisterPackage(FlipCommodity, "FlipCommodity", 1, {})
