-----------------------------------------------------------------------------------------------
--[[ Copyright (c) chckxy (watdametz@gmail.com). All rights reserved

This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.  --]]
-----------------------------------------------------------------------------------------------
 
local FlipCrafting = {} 
local FlipCraftHeader = {}

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipCrafting:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
    return o
end

function FlipCrafting:Init()
	-- a table of the craft headers
	self.tHeaders = {}
	self.nMinFontSize = self.nMinFontSize 
	
	-- have the market instance
	self.market = self.market
	
	-- setup window
	self.xmlDoc = XmlDoc.CreateFromFile("Flip.xml")	
	self.xmlGroupDoc = XmlDoc.CreateFromFile("FlipGroups.xml")
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "FlipForm", nil, self)
	self.wndLeft = self.wndMain:FindChild("Left")
	self.wndRight = self.wndMain:FindChild("Right")

	self.wndTabs = self.wndMain:FindChild("TabContainer")
	local tab = Apollo.LoadForm(self.xmlDoc, "Tab", self.wndTabs, self)
	tab:SetText("Crafting")

	self.tCores = {
		[CraftingLib.CodeEnumTradeskillTier.Novice] 	= {},
		[CraftingLib.CodeEnumTradeskillTier.Apprentice] = {},
		[CraftingLib.CodeEnumTradeskillTier.Journeyman] = {},
		[CraftingLib.CodeEnumTradeskillTier.Artisan] 	= {},
		[CraftingLib.CodeEnumTradeskillTier.Expert] 	= {},
	}
	
	self.timerDelay = Apollo.CreateTimer("CraftTimerDelay", 2, false)
	Apollo.StopTimer("CraftTimerDelay")
	Apollo.RegisterTimerHandler("CraftTimerDelay", "Redraw", self)
end

function FlipCrafting:OnFlipCraftOn()
	-- if this is the first time we've ever loaded
	if not self.tTradeskills then self:LoadSchematics() end
	self.wndMain:Invoke()
end

function FlipCrafting:Dirty()
	Apollo.StopTimer("CraftTimerDelay")
	Apollo.StartTimer("CraftTimerDelay")
end

function FlipCrafting:ToSave()
	local tCores = {}
	for nTier, t in ipairs(self.tCores) do
		local tSave = {}
		tCores[nTier] = tSave
		for nId, item in pairs(t) do
			table.insert(tSave, nId)
		end
	end
	return { tCores = tCores, }
end

function FlipCrafting:RestoreSave( tCraftSave )
	for nTier, tIds in ipairs(tCraftSave.tCores or {}) do
		local tCores = self.tCores[nTier]
		for nKey, nId in ipairs(tIds) do
			tCores[nId] = Item.GetDataFromId(nId)
		end
	end
	return true
end

function FlipCrafting:SetMinFontSize( nSize )
	self.nMinFontSize = nSize
end

function FlipCrafting:Redraw()
	for nId, header in pairs(self.tHeaders) do
		header:Redraw()--self:GetFilterAmount())
	end
	self:Reorder()
end

function FlipCrafting:Reorder()
	self.wndRight:ArrangeChildrenVert(0, function(wnd1, wnd2)
		local n1 = wnd1:GetData()
		local n2 = wnd2:GetData()
		return n1 > n2
	end)
end

function FlipCrafting:LoadSchematics()
	-- copied from TradeskillSchematics
	self.tTradeskills = {}
	for idx, tCurrTradeskill in pairs(CraftingLib.GetKnownTradeskills()) do
		local tCurrTradeskillInfo = CraftingLib.GetTradeskillInfo(tCurrTradeskill.eId)
		if not tCurrTradeskillInfo.bIsHarvesting and tCurrTradeskill.eId ~= CraftingLib.CodeEnumTradeskill.Runecrafting then
			table.insert(self.tTradeskills, { 
				tSkill = tCurrTradeskill, 
				tInfo = tCurrTradeskillInfo,
				tSchematics = CraftingLib.GetSchematicList(tCurrTradeskill.eId, nil, nil, true)
			})
		end
	end

	for idx, tCurrData in ipairs(self.tTradeskills) do
		for idx, tSchematic in ipairs(tCurrData.tSchematics) do
			if tSchematic.bIsKnown then
				local tHeader = FlipCraftHeader:new({ tSchematic = tSchematic, crafting = self })
				tHeader:Redraw()
				local tSchematicInfo = CraftingLib.GetSchematicInfo(tSchematic.nSchematicId)
				local nResultId = tSchematicInfo.itemOutput:GetItemId()
				table.insert(self.tHeaders, tHeader)
			end
		end
	end
	
	--self:Redraw()
	self:Reorder()
end

function FlipCrafting:AddWindow( strName, wndParent, obj )
	return Apollo.LoadForm(self.xmlDoc, strName, wndParent, obj)
end

function FlipCrafting:AddHeaderWindow( obj )
	local wnd = Apollo.LoadForm(self.xmlGroupDoc, "ItemHeader", self.wndRight, obj)
	Apollo.LoadForm(self.xmlGroupDoc, "ItemPictureContainer", wnd, obj)
	return wnd 
end

function FlipCrafting:AddSub( nId, obj )
	self.market:AddSub(nId, obj)
end

function FlipCrafting:RemoveSub( nId, obj )
	self.market:RemoveSub(nId, obj)
end

function FlipCrafting:GetLowestBuy( nId, nAmount )
	local strName, nBuy = self.market:GetLowestBuy(nId, nAmount)
	return strName, nBuy
end

function FlipCrafting:GetHighestSell( nId, nAmount )
	local strName, nSell = self.market:GetHighestSell(nId, nAmount)
	return strName, nSell
end

-- GetFilterAmount : -> Number
-- get the value of the filter slider
function FlipCrafting:GetFilterAmount()
end

-- Number -> Number
-- set the value of the filter slider and return it
function FlipCrafting:SetFilterAmount(nFilterAmount)
end

-- OnFilterChange : FilterWindow Window Float Float -> 
function FlipCrafting:OnFilterChange( wndHandler, wndControl, fNewValue, fOldValue )
end

function FlipCrafting:DisplayOptions( wndParent, group )
end

function FlipCrafting:OnClose()
	self.tTradeskills = nil
	for nKey, header in ipairs(self.tHeaders) do
		header:Destroy()
	end
	self.tHeaders = {}
	self.wndMain:Close()
end

function FlipCrafting:AddCore(nTier, item)
	local tCores = self.tCores[nTier]
	tCores[item:GetItemId()] = item
end

function FlipCrafting:GetCores(nTier)
	return self.tCores[nTier]
end

------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------


function FlipCraftHeader:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
	return o
end

function FlipCraftHeader:Init()
	self.crafting = self.crafting
	
	-- information about the schematic we're crafting
	self.tSchematic = self.tSchematic
	self.tSchematicInfo = CraftingLib.GetSchematicInfo(self.tSchematic.nSchematicId)
	
	-- information about the output item
	self.itemData = self.tSchematicInfo.itemOutput

	-- make our window
	self.wnd = self.crafting:AddHeaderWindow(self)
	self.wnd:FindChild("ItemName"):SetText(self.itemData:GetName())

	-- initialize the window
	local container = self.crafting:AddWindow("Container", self.wnd, self)
	self.wndCostContainer = self.crafting:AddWindow("CraftContainer", container, self)
	self.wndProfitContainer = self.crafting:AddWindow("CraftContainer", container, self)
	container:ArrangeChildrenHorz(0)

	-- make the picture pretty
	local wndItemPicture = self.wnd:FindChild("ItemPicture")
	wndItemPicture:SetSprite(self.itemData:GetIcon())
	wndItemPicture:SetData(self.itemData)

	-- set the materials we need for the craft
	self.tMaterials = self.tSchematicInfo.tMaterials

	-- set our cores (this isn't as simple as just getting the available cores,
	-- as we might have save data with cores that aren't currently available)
	self.tCores = nil

	for key, tMaterial in pairs(self.tMaterials) do
		local nId = tMaterial.itemMaterial:GetItemId()
		self.crafting:AddSub(nId, self)
	end

	local tAvailableCores = CraftingLib.GetAvailablePowerCores(self.tSchematic.nSchematicId)
	if tAvailableCores then -- some crafts won't have power cores
		local eTier = self.tSchematicInfo.eTier
		for idx, itemCore in pairs(tAvailableCores) do
			self.crafting:AddCore(eTier, itemCore)
		end

		self.tCores = self.crafting:GetCores(eTier) or {}
		for nId, itemCore in pairs(self.tCores) do
			self.crafting:AddSub(nId, self)
		end
	end

	self.crafting:AddSub(self.itemData:GetItemId(), self)
end

-----------------------------------------------------------------------------------------------
-- generate the tooltip for headers
-- copied from MarketplaceCommodity
function FlipCraftHeader:OnGenerateTooltipListIcon( wndHandler, wndControl, eType, nX, nY )
	local tCurrItem = wndHandler:GetData()
	if tCurrItem then -- moving the mouse around a lot throws errors here....
		Tooltip.GetItemTooltipForm(self, wndHandler, tCurrItem, {itemCompare = tCurrItem:GetEquippedItemForItemType()})
	end
end

------------------------------------------------------------------------------------------------------------

function FlipCraftHeader:Notify( nId, tPrice )
	self.crafting:Dirty()
	-- self:Redraw()--self.crafting:GetFilterAmount())
end

function FlipCraftHeader:Redraw()-- nFilterAmount )
	-- we have updated material prices, so calculate the profit of crafting
	-- vs the profit of just flipping the materials, and the cost of buying those materials
	local nMatBuyPrice = 0
	local strMatBuy = ""
	local nMatSellPrice = 0
	local strMatBuy = ""
	for nKey, tMaterial in pairs(self.tMaterials) do
		local itemMaterial = tMaterial.itemMaterial
		local nAmount = tMaterial.nAmount

		local strBuy, nBuy = self.crafting:GetLowestBuy(itemMaterial:GetItemId(), nAmount)
		local strSell, nSell = self.crafting:GetHighestSell(itemMaterial:GetItemId(), nAmount)
		
		if nBuy and nMatBuyPrice and nMatSellPrice then
			nMatBuyPrice = nMatBuyPrice + (nBuy * nAmount)
			strMatBuy = strMatBuy .. "\n" .. strBuy
			nMatSellPrice = nMatSellPrice + (math.max(nBuy, nSell or 0) * nAmount)
		elseif not nBuy then
			strMatBuy = strMatBuy .. "\n" .. itemMaterial:GetName() .. " MISSING PRICE"
			nMatBuyPrice = nil
			nMatSellPrice = nil
		end
	end

	-- calculate the core cost
	local nCoreBuy = nil
	local strCoreBuy = ""
	local nCoreSell = 0
	local strCoreSell = ""
	local nAmount = 0

	if self.tCores then
		-- pick the core with the lowest buy cost? not sure how to do this right
		-- TODO: add preferences for core selection?
		for idx, itemMaterial in pairs(self.tCores) do
			nAmount = 1
	
			local strBuy, nBuy = self.crafting:GetLowestBuy(itemMaterial:GetItemId(), nAmount)
			local strSell, nSell = self.crafting:GetHighestSell(itemMaterial:GetItemId(), nAmount)
			
			if (not nCoreBuy) or (nBuy and nBuy < nCoreBuy) then
				nCoreBuy = nBuy
				strCoreBuy = strBuy
				nCoreSell = math.max(nBuy or 0, nSell or 0)
				strCoreSell = strSell
			end
		end
		if nAmount == 0 or not nCoreBuy then
			strCoreBuy = "MISSING CORE PRICE"
			nCoreBuy = nil
			nCoreSell = nil
		end
	end

	local nTotalBuy = (nMatBuyPrice and (nMatBuyPrice + (nCoreBuy or 0))) or 0
	self.wndCostContainer:FindChild("Price"):SetAmount(nTotalBuy)
	self.wndCostContainer:FindChild("Price"):SetTextColor(nTotalBuy > 0 and "white" or "red")
	self.wndCostContainer:SetTooltip(strMatBuy .. "\n" .. strCoreBuy)

	local strOutputSell, nOutputSell = self.crafting:GetHighestSell(self.itemData:GetItemId(), 1)
	if nOutputSell and nMatSellPrice and nCoreSell then
		local nTotalSell = nMatSellPrice + nCoreSell
		self.nProfit = nOutputSell - nTotalSell
		self.wndProfitContainer:FindChild("Price"):SetAmount(math.abs(self.nProfit))
		self.wndProfitContainer:SetTooltip(strOutputSell)

		local bProfit = self.nProfit > 0
		self.wndProfitContainer:FindChild("Price"):SetTextColor(bProfit and "white" or "red")		
	elseif not nOutputSell then
		self.wndProfitContainer:FindChild("Price"):SetTextColor("red")
		self.wndProfitContainer:SetTooltip(self.itemData:GetName() .. " MISSING PRICE")
	else
		self.wndProfitContainer:FindChild("Price"):SetTextColor("red")
		self.wndProfitContainer:SetTooltip("Cannot calculate profit with missing material prices")		
	end
	
	self.wnd:SetData(self.nProfit or 0)
	self.crafting:Reorder()
end

function FlipCraftHeader:Destroy()
	self.wnd:Destroy()
	self.crafting:RemoveSub(self.itemData:GetItemId(), self)
end

------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------

Apollo.RegisterPackage(FlipCrafting, "FlipCrafting", 1, {})
