-----------------------------------------------------------------------------------------------
--[[ Copyright (c) chckxy (watdametz@gmail.com). All rights reserved

This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.  --]]
-----------------------------------------------------------------------------------------------

local FlipMarket = {}
local FlipMarketOptions = {}
local FlipMarketOption = {}

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipMarket:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o:Init()
    return o
end

function FlipMarket:Init()
	-- initial fields
	self.flip = self.flip
	self.groups = self.groups
	self.nMinFontSize = self.nMinFontSize
	self.bSplitCancel = true
	self.bOneClickDelete = false
	self.bOnMarketplace = false
	self.bOnAuction = false
	self.bHideDefaultPostMessage = true

	-- we have headers
	self.tHeaders = {}

	-- we keep track of all market prices here
	self.tPricesCX = {}
	self.tPricesAH = {}
	self.tSubs = {} -- things with Notify methods for price changes

	-- setup the main window
	self.xmlDoc = XmlDoc.CreateFromFile("Flip.xml")	
	self.xmlOptionDoc = XmlDoc.CreateFromFile("FlipGroups.xml")	
	self.wndMain = Apollo.LoadForm(self.xmlDoc, "FlipForm", nil, self)
    self.wndMain:Show(false, true)

	-- controls the CX side of the "/flip" window
	self.commodity = Apollo.GetPackage("FlipCommodity").tPackage:new({ 
		market = self, 
		groups = self.groups,
		wndMain = self.wndMain,
	})
	
	self.auction = Apollo.GetPackage("FlipAuction").tPackage:new({
		market = self,
		groups = self.groups,
		wndMain = self.wndMain,
	})

	-- we have options
	self.options = FlipMarketOptions:new({ xmlDoc = self.xmlOptionDoc })

	-- there are many tabs but this is our tab
	self.tab = self.groups:AddTab(self)

	-- the height of an options row, for the group dropdown menu option
	local wndRow = Apollo.LoadForm(self.xmlOptionDoc, "GroupOptionsRow", wndParent, self)
	self.nRowHeight = wndRow:GetHeight()
	
	self.ma = Apollo.GetAddon("MarketplaceAuction")
	self.timerDirty = Apollo.CreateTimer("AuctionTimerDirty", 1, false)
	Apollo.StopTimer("AuctionTimerDirty")
	Apollo.RegisterTimerHandler("AuctionTimerDirty", "RestoreMarketplaceAuction", self)
end

-----------------------------------------------------------------------------------------------
-- slash commands

function FlipMarket:OnFlipOn()
	self.auction:OnClose()
	self.commodity:Open()
end

function FlipMarket:OnAuctionOn()
	self.commodity:OnClose()
	self.auction:Open()
end

function FlipMarket:OnFlipMarketOn()
	self.groups:OnTab(self.tab)
end

-- OnClose : ->
-- on the close button, close the window!
function FlipMarket:OnClose()
	self.wndMain:Close()
end

-- the window for market options
function FlipMarket:Open(wnd, wndLeft, wndRight)
	--local wndLeft = self.wndMarket:FindChild("GroupHeadersFull")
	wndLeft:DestroyChildren()
	wndRight:DestroyChildren()
	self.options:DisplayOptionDescriptions(wndLeft, wndRight)
	self.wndRight = wndRight
	wndLeft:ArrangeChildrenVert(0)
	wnd:Invoke()
end

function FlipMarket:TabName() 
	return "Market"
end

-----------------------------------------------------------------------------------------------
-- configuration

-- the flip module redirects configuration here; this loads the window
function FlipMarket:OnConfigure()
	self.wndConfigure = Apollo.LoadForm(self.xmlOptionDoc, "MarketplaceOptions", nil, self)
	self.wndConfigure:FindChild("OnMarketplaceOpen"):SetCheck(self.bOnMarketplace)
	self.wndConfigure:FindChild("OnAuctionOpen"):SetCheck(self.bOnAuction)
	self.wndConfigure:FindChild("SplitCancelUndercut"):SetCheck(self.bSplitCancel)
	self.wndConfigure:FindChild("OneClickDelete"):SetCheck(self.bOneClickDelete)
	self.wndConfigure:FindChild("PostMessage"):SetCheck(self.bHideDefaultPostMessage)
	self.wndConfigure:FindChild("MinFontSlider"):SetValue(self.nMinFontSize or 10)
end

-- OnMarketplaceOpenToggle : ->
-- toggling whether the /flip window should appear when the CX opens
function FlipMarket:OnMarketplaceOpenToggle()
	self.bOnMarketplace = not self.bOnMarketplace
end

function FlipMarket:OnAuctionOpenToggle()
	self.bOnAuction = not self.bOnAuction
end

function FlipMarket:OnSplitCancelToggle()
	self.bSplitCancel = not self.bSplitCancel
end

function FlipMarket:OnOneClickDelete()
	self.bOneClickDelete = not self.bOneClickDelete
end

function FlipMarket:OnPostMessageToggle()
	self.bHideDefaultPostMessage = not self.bHideDefaultPostMessage
	self.commodity:HidePostMessageToggle(self.bHideDefaultPostMessage)
end

-- OnSaveButtonSignal : SaveButton -> 
-- saving the configurations
function FlipMarket:OnSaveButtonSignal( btnSave )
	btnSave:GetParent():Close()
end

function FlipMarket:OnMinFontSizeChange( wndHandler, wndControl, fNewValue, fOldValue )
	if fNewValue ~= fOldValue then self.flip:SetMinFontSize(fNewValue) end
end

function FlipMarket:SetMinFontSize( nSize )
	self.nMinFontSize = nSize
end

function FlipMarket:SplitCancel()
	return self.bSplitCancel
end

function FlipMarket:OneClickDelete()
	return self.bOneClickDelete
end

-----------------------------------------------------------------------------------------------
-- saving

function FlipMarket:RestoreSave( tMarketSave )
	if tMarketSave.tCommoditySave then self.commodity:RestoreSave(tMarketSave.tCommoditySave) end
	if tMarketSave.tOptions then self.options:RestoreSave(tMarketSave.tOptions) end
	if tMarketSave.bOnMarketplace then self.bOnMarketplace = true end
	if tMarketSave.bOnAuction then self.bOnAuction = true end
	if tMarketSave.bSplitCancel == false then self.bSplitCancel = false end
	if tMarketSave.bOneClickDelete then self.bOneClickDelete = true end
	if tMarketSave.bHideDefaultPostMessage == false then
		self.bHideDefaultPostMessage = false
		self.commodity:HidePostMessageToggle(false)
	end

	for nId, tSave in pairs(tMarketSave.tPricesCX or tMarketSave.tPrices or tMarketSave or {}) do
		if type(nId) == "number" then
			local tPrice = self.tPricesCX[nId]
			if not tPrice then
				tPrice = {}
				table.insert(self.tPricesCX, nId, tPrice)
			end
			tPrice.nBuy1Price = tSave.nBuy1Price or 0
			tPrice.nBuy10Price = tSave.nBuy10Price or 0
			tPrice.nBuy50Price = tSave.nBuy50Price or 0
			tPrice.nSell1Price = tSave.nSell1Price or 0
			tPrice.nSell10Price = tSave.nSell10Price or 0
			tPrice.nSell50Price = tSave.nSell50Price or 0
			tPrice.nBuyVendor = tSave.nBuyVendor or 0
			local cSellVendor = Item.GetDataFromId(nId):GetSellPrice()
			tPrice.nSellVendor = tSave.nSellVendor or cSellVendor and cSellVendor:GetAmount() or 0
		end
	end

	for nId, tSave in pairs(tMarketSave.tPricesAH or {}) do
		local tPrice = self.tPricesAH[nId]
		if not tPrice then
			tPrice = {}
			table.insert(self.tPricesAH, nId, tPrice)
		end
		tPrice.nBuyout = tSave.nBuyout or 0
		tPrice.nBuyVendor = tSave.nBuyVendor or 0
		tPrice.nSellVendor = tSave.nSellVendor or Item.GetDataFromId(nId):GetSellPrice():GetAmount() or 0
	end
	
	if tMarketSave.tWindowPoints then
		self.wndMain:SetAnchorPoints(unpack(tMarketSave.tWindowPoints))
	end
	
	if tMarketSave.tWindowOffsets then
		self.wndMain:SetAnchorOffsets(unpack(tMarketSave.tWindowOffsets))
	end
	
	return true
end

function FlipMarket:ToSave()
	return {
		tWindowPoints = { self.wndMain:GetAnchorPoints() },
		tWindowOffsets = { self.wndMain:GetAnchorOffsets() },
		tCommoditySave = self.commodity:ToSave(),
		tPricesCX = self.tPricesCX,
		tPricesAH = self.tPricesAH,
		tOptions = self.options:ToSave(),
		bOnMarketplace = self.bOnMarketplace,
		bOnAuction = self.bOnAuction,
		bSplitCancel = self.bSplitCancel,
		bOneClickDelete = self.bOneClickDelete,
		bHideDefaultPostMessage = self.bHideDefaultPostMessage,
	}
end

-----------------------------------------------------------------------------------------------

-- AddSub : ItemID Object -> 
-- adds object as a subscriber to changes in itemID
-- TODO: do this with generic events instead? not sure if faster, probably simpler for us
function FlipMarket:AddSub( nId, obj )
	local tSub = self.tSubs[nId]
	if tSub then
		table.insert(tSub, obj)
	else
		table.insert(self.tSubs, nId, { obj })
	end

	local tPrice = self:GetPrice(nId)
	if not tPrice then
		local cSellVendor = Item.GetDataFromId(nId):GetSellPrice()
		tPrice = { nSellVendor = cSellVendor and cSellVendor:GetAmount() or 0, }
		local bIsAH = Item.GetDataFromId(nId):IsAuctionable()
		table.insert(bIsAH and self.tPricesAH or self.tPricesCX, nId, tPrice)
	end
end

function FlipMarket:RemoveSub( nId, obj )
	local tSub = self.tSubs[nId]
	if tSub then
		for nKey, sub in ipairs(tSub) do
			if sub == obj then 
				table.remove(tSub, nKey) 
				return 
			end
		end
	end
end

-- GetPrice : ItemID -> PriceTable
-- return the stored price table for nId
function FlipMarket:GetPrice( nId )
	return self.tPricesCX[nId] or self.tPricesAH[nId]
end

-- GetLowestBuy : ItemID Number -> String Number
-- returns a string describing the other return value, the lowest price for buying the given item id
function FlipMarket:GetLowestBuy( nId, nAmount )
	local tPrice = self:GetPrice(nId)
	
	-- possible price sources
	local nBuyCX = nil
	local nBuyVendor = nil
	local nBuyAH = nil
	
	-- if it's a CX item, get the price for the correct amount
	-- TODO: this won't work for materials with a small sample size
	if tPrice.nBuy50Price  then
		if     nAmount <= 1  then nBuyCX = tPrice.nBuy1Price > 0  and tPrice.nBuy1Price
		elseif nAmount <= 10 then nBuyCX = tPrice.nBuy10Price > 0 and tPrice.nBuy10Price
		else                      nBuyCX = tPrice.nBuy50Price > 0 and tPrice.nBuy50Price end
	end
	
	-- if we can buy it on the CX thru a sell order, try that too
	if not nBuyCX and tPrice.nSell50Price then
		if     nAmount <= 1  then nBuyCX = tPrice.nSell1Price > 0  and tPrice.nSell1Price
		elseif nAmount <= 10 then nBuyCX = tPrice.nSell10Price > 0 and tPrice.nSell10Price
		else                      nBuyCX = tPrice.nSell50Price > 0 and tPrice.nSell50Price end
	end

	-- if we can buy it from a vendor, get the price
	if tPrice.nBuyVendor then nBuyVendor = tPrice.nBuyVendor > 0 and tPrice.nBuyVendor end
	
	-- if we can buy it from the AH, get the price
	if tPrice.nBuyout then nBuyAH = tPrice.nBuyout > 0 and tPrice.nBuyout end
	
	local tPrices = {
		CX = nBuyCX,
		Vendor = nBuyVendor,
		AH = nBuyAH,
	}
	
	local nMin = nil
	local str = ""
	
	for strName, nPrice in pairs(tPrices) do
		if (not nMin) or (nPrice and nPrice < nMin) then 
			nMin = nPrice 
		end
		if nPrice then
			-- TODO: learn how LUA format strings work lol
			str = str .. "Buy " .. Item.GetDataFromId(nId):GetName() .. " x" .. nAmount .. " from " .. strName .. " for " .. nPrice .. "\n"
		end
	end

	return str, nMin
end

-- GetHighestSell : ItemID Number -> String Number
-- returns a string describing how to get the most out of the given item, also returns the highest sell price
function FlipMarket:GetHighestSell( nId, nAmount )
	local tPrice = self:GetPrice(nId)
	local itemData = Item.GetDataFromId(nId)

	-- possible price sources
	local nSellCX = nil
	local nSellVendor = nil
	local nSellAH = nil

	-- if it's a CX item, get the price for the correct amount
	-- TODO: doesn't work good for small samples
	if tPrice.nSell50Price  then
		if     nAmount <= 1  then nSellCX = tPrice.nSell1Price > 0  and tPrice.nSell1Price
		elseif nAmount <= 10 then nSellCX = tPrice.nSell10Price > 0 and tPrice.nSell10Price
		else                      nSellCX = tPrice.nSell50Price > 0 and tPrice.nSell50Price end
		-- nSellCX = nSellCX and nSellCX * 0.88
	end

	-- if we can buy it from a vendor, get the price
	if tPrice.nSellVendor then nSellVendor = tPrice.nSellVendor > 0 and tPrice.nSellVendor end

	-- if we can buy it from the AH, get the price
	if tPrice.nBuyout then nSellAH = tPrice.nBuyout > 0 and tPrice.nBuyout end

	local tPrices = {
		CX = nSellCX,
		Vendor = nSellVendor,
		AH = nSellAH,
	}

	local nMax = nil
	local str = ""

	for strName, nPrice in pairs(tPrices) do
		if (not nMax) or (nPrice and nPrice > nMax) then 
			nMax = nPrice 
			strMax = strName
		end
		if nPrice then
			-- TODO: LUA format string
			str = str .. "Sell " .. itemData:GetName() .. " x" .. nAmount .. " to " .. strName .. " for " .. nPrice .. "\n"
		end
	end

	return str, nMax
end

-----------------------------------------------------------------------------------------------
-- dealing with commodity headers

-- AddListItem : Window -> wndListItem
-- adds a ListItem to wndParent
function FlipMarket:AddWindow( strName, wndParent, objHandler )
	local wnd = Apollo.LoadForm(self.xmlDoc, strName, wndParent, objHandler)
	for nKey, strName in ipairs({ "Price", "Quantity" }) do
		local maybeText = wnd:FindChild(strName)
		if maybeText then maybeText:SetFont("CRB_Interface" .. self.nMinFontSize) end
	end
	return wnd
end


-----------------------------------------------------------------------------------------------
-- getting updated info from market

-- when marketplace opens
function FlipMarket:OnToggleMarketplaceWindow()
	if self.bOnMarketplace then self:OnFlipOn() end
	self:RefreshCX()
end

function FlipMarket:OnToggleAuctionWindow()
	if self.bOnAuction then self:OnAuctionOn() end
	-- self:RefreshAH()
end

function FlipMarket:NotifySubs( nId, tPrice )
	for nKey, obj in ipairs(self.tSubs[nId] or {}) do
		obj:Notify(nId, tPrice)
	end
end

-- CommodityID CommodityStats ? -> 
function FlipMarket:OnCommodityInfoResults( nId, tStats, tOrders )
	local tPrice = self.tPricesCX[nId]
	if tPrice then
		tPrice.nBuy1Price = tStats.arBuyOrderPrices[1].monPrice:GetAmount() or 0
		tPrice.nBuy10Price = tStats.arBuyOrderPrices[2].monPrice:GetAmount() or 0
		tPrice.nBuy50Price = tStats.arBuyOrderPrices[3].monPrice:GetAmount() or 0
		tPrice.nSell1Price = tStats.arSellOrderPrices[1].monPrice:GetAmount() or 0
		tPrice.nSell10Price = tStats.arSellOrderPrices[2].monPrice:GetAmount() or 0
		tPrice.nSell50Price = tStats.arSellOrderPrices[3].monPrice:GetAmount() or 0

		self:NotifySubs(nId, tPrice)
	end
end

function FlipMarket:OnItemAuctionSearchResults(nPage, nTotalResults, tAuctions)
	self.nCurPage = nPage
	for idx, aucCurr in ipairs(tAuctions) do
		local itemData = aucCurr:GetItem()
		if not aucCurr:IsOwned() then
			local nBuyoutPrice = aucCurr:GetBuyoutPrice():GetAmount()
			local nId = itemData:GetItemId()
			local tPrice = self.tPricesAH[nId]
			if tPrice and nBuyoutPrice > 0 and not tPrice.nBuyout then
				tPrice.nBuyout = nBuyoutPrice
				self:NotifySubs(nId, tPrice)
				break
			end
		end	
	end
	Apollo.StopTimer("AuctionTimerDirty")
	Apollo.StartTimer("AuctionTimerDirty")
end

function FlipMarket:OnInvokeVendorWindow( unitArg )
	local tVendorList = unitArg:GetVendorItems()
	if tVendorList then 
		for key,value in ipairs(tVendorList) do
			local itemData = value.itemData
			local nId = itemData:GetItemId()
			local tPrice = self.tPricesCX[nId] or self.tPricesAH[nId]
			if tPrice then
				local nOldPrice = tPrice.nBuyVendor 
				tPrice.nBuyVendor = itemData:GetBuyPrice():GetAmount()
				if nOldPrice ~= tPrice.nBuyVendor then
					self:NotifySubs(nId, tPrice)
				end
			end
		end
	end
end

function FlipMarket:RefreshAH()
	Apollo.StopTimer("AuctionTimerDirty")
	Apollo.RemoveEventHandler("ItemAuctionSearchResults", self.ma)
	local nPage = 0
	local eAuctionSort = MarketplaceLib.AuctionSort.Buyout
	local bReverseSort = false
	local arFilters = {}
	local nPropertySort = false
	for nId, tPrice in pairs(self.tPricesAH) do
		tPrice.nBuyout = nil
		MarketplaceLib.RequestItemAuctionsByItems({ nId }, nPage, eAuctionSort, bReverseSort, arFilters, nil, nil, nPropertySort)
	end
end

function FlipMarket:RestoreMarketplaceAuction()
	Apollo.RegisterEventHandler("ItemAuctionSearchResults", "OnItemAuctionSearchResults", self.ma)
end

function FlipMarket:RefreshCX()
	for nId, tPrice in pairs(self.tPricesCX) do
		MarketplaceLib.RequestCommodityInfo(nId)
	end
	MarketplaceLib.RequestOwnedCommodityOrders()
end

function FlipMarket:GetOptions( group, tOptionSave )
	return FlipMarketOption:new({ 
		group = group, 
		options = self.options:GetTable(),
		tSave = tOptionSave,
	})
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipMarketOptions:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

-- the window ordering
FlipMarketOptions.Ordering = {
	"BUYOPTION", "SELLOPTION", "COLOROPTION",
}

-- these will always be numbers, so you're safe adding/subtracting
FlipMarketOptions.Price1Options = {
	BUYTOP1 = 1,
	BUYTOP10 = 2,
	BUYTOP50 = 3,
	SELLTOP1 = 4,
	SELLTOP10 = 5,
	SELLTOP50 = 6,
	ALWAYS = 7,
	NEVER = 8,
}

-- these will always be numbers, so you're safe adding/subtracting
FlipMarketOptions.Price2Options = {
	BUYTOP1 = 1,
	BUYTOP10 = 2,
	BUYTOP50 = 3,
	SELLTOP1 = 4,
	SELLTOP10 = 5,
	SELLTOP50 = 6,
}

FlipMarketOptions.PriceOptionsText = {
	[FlipMarketOptions.Price1Options.BUYTOP1] = "Buy Top1",
	[FlipMarketOptions.Price1Options.BUYTOP10] = "Buy Top10",
	[FlipMarketOptions.Price1Options.BUYTOP50] = "Buy Top50",
	[FlipMarketOptions.Price1Options.SELLTOP1] = "Sell Top1",
	[FlipMarketOptions.Price1Options.SELLTOP10] = "Sell Top10",
	[FlipMarketOptions.Price1Options.SELLTOP50] = "Sell Top50",
	[FlipMarketOptions.Price1Options.ALWAYS] = "Always",
	[FlipMarketOptions.Price1Options.NEVER] = "Never",
}

FlipMarketOptions.PriceComparisons = {
	LESS = 1,
	GREATER = 2,
}

FlipMarketOptions.PriceOffset = {
	CONSTANT = 1,
	PERCENT = 2,
}

FlipMarketOptions.Quantity = {
	CONSTANT = 1,
	ALL = 2,
}

function FlipMarketOptions:Init()
	self.options = {
		--[FlipMarketOptions.Categories.BUYCONDITION] =
		--	{ strDescription = "When the following is true:", tOptions = {} },
		BUYOPTION = { strDescription = "Post the following buy order:", tOptions = {} },
		--[FlipMarketOptions.Categories.SELLCONDITION] =
		--	{ strDescription = "When the following is true:", tOptions = {} },
		SELLOPTION = { strDescription = "Post the following sell order:", tOptions = {} },
		COLOROPTION = { strDescription = "Coloring:", tOptions = {} },
	}

	self.xmlDoc = self.xmlDoc
	self:SetupInitialOptions()
end

-- make sure everything is in the structure we're expecting
FlipMarketOptions.Empty = {
	--[[BUYCONDITION = {
		strName = "",
		ePrice1 = FlipMarketOptions.Price1Options.BUYTOP1,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.LESS,
		ePrice2 = FlipMarketOptions.Price2Options.SELLTOP1,
	},--]]
	BUYOPTION = {
		strName = "",
		eQuantity = FlipMarketOptions.Quantity.CONSTANT,
		nQuantity = 200,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.GREATER,
		ePrice2 = FlipMarketOptions.Price2Options.SELLTOP1
	},
	--[[SELLCONDITION = {
		strName = "",
		ePrice1 = FlipMarketOptions.Price1Options.ALWAYS,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.GREATER,
		ePrice2 = FlipMarketOptions.Price2Options.BUYTOP1,
	},--]]
	SELLOPTION = {
		strName = "",
		eQuantity = FlipMarketOptions.Quantity.ALLIN,
		nQuantity = 0,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT, 
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.LESS,
		ePrice2 = FlipMarketOptions.Price2Options.SELLTOP1,
	},
	COLOROPTION = { 
		strName = "New Color Option", 
		tFixed = { ORANGE = 0, YELLOW = 0, GREEN = 0, }, 
		tPercent = { ORANGE = 0, YELLOW = 0, GREEN = 0, }, 
		tColors = { RED = "red", ORANGE = "xkcdOrangePink", YELLOW = "xkcdAmber", GREEN = "xkcdAlgaeGreen" },
 	},
}

-----------------------------------------------------------------------------------------------

function FlipMarketOptions:ToSave()
	return self.options
end

function FlipMarketOptions:RestoreSave( options )
	for eType, value in pairs(self.options) do
		if options[eType] then 
			self.options[eType] = options[eType]
			self:CheckOption(eType)
		end
	end
end

function FlipMarketOptions:CheckOption( eType )
	local empty = FlipMarketOptions.Empty[eType]
	for nKey, option in ipairs(self.options[eType].tOptions) do
		for eKey, value in pairs(empty) do
			if not option[eKey] then option[eKey] = value end
		end
		for eKey, value in pairs(option) do
			if not empty[eKey] then option[eKey] = nil end
		end
	end
end

-----------------------------------------------------------------------------------------------

function FlipMarketOptions:GetTable()
	return self.options
end

function FlipMarketOptions:GetOptionTableElement(eType, nKey)
end

-----------------------------------------------------------------------------------------------
-- /flipmarket 

function FlipMarketOptions:DisplayOptionDescriptions( wndLeft, wndRight )
	self.wndLeft = wndLeft
	self.wndRight = wndRight
	for nKey, eType in ipairs(FlipMarketOptions.Ordering) do
		local wnd = Apollo.LoadForm(self.xmlDoc, "GroupHeader", wndLeft, self)
		wnd:FindChild("GroupName"):SetText(self.options[eType].strDescription)
		wnd:SetData(eType)
	end
end

function FlipMarketOptions:OnGroupHeaderButtonDown( wndHandler )
	local eType = wndHandler:GetParent():GetData()
	self.wndRight:SetData(eType)
	self.wndRight:DestroyChildren()
	self:RenderOptions(eType, self.wndRight)
end

-- when you hit the + button in a flipmarket option
function FlipMarketOptions:OnAddNewOption( wndHandler, wndControl, eMouseButton )
	local eType = self.wndRight:GetData()
	local optionNew = {}
	for eKey, value in pairs(FlipMarketOptions.Empty[eType]) do
		optionNew[eKey] = value
	end
	self:AddOption(eType, optionNew)
	self.wndRight:DestroyChildren()
	self:RenderOptions(eType, self.wndRight)
end

function FlipMarketOptions:OnDeleteOption( wndHandler, wndControl, eMouseButton )
	local eType = self.wndRight:GetData()
	local option = wndHandler:GetParent():GetData()
	if self:RemoveOption(eType, option) then
		self.wndRight:DestroyChildren()
		self:RenderOptions(eType, self.wndRight)
	end
end

-----------------------------------------------------------------------------------------------
-- rendering order options

function FlipMarketOptions:RenderBuyOptions( options, wndParent )
	for nKey, option in ipairs(options) do
		local wnd = Apollo.LoadForm(self.xmlDoc, "MarketplaceOrderOptions", wndParent, self)
		wnd:FindChild("OptionName"):SetText(option.strName)
		wnd:SetData(option)

		wnd:FindChild("TogglePrice1"):Show(false)
		self:RenderQuantity(option, wnd)
		self:RenderOffset(option, wnd)
		self:RenderRelation(option, wnd)
		self:RenderPrice2(option, wnd)
	end
end

function FlipMarketOptions:RenderSellOptions( options, wndParent )
	for nKey, option in ipairs(options) do
		local wnd = Apollo.LoadForm(self.xmlDoc, "MarketplaceOrderOptions", wndParent, self)
		wnd:FindChild("OptionName"):SetText(option.strName)
		wnd:SetData(option)

		wnd:FindChild("TogglePrice1"):Show(false)
		self:RenderQuantity(option, wnd)
		self:RenderOffset(option, wnd)
		self:RenderRelation(option, wnd)
		self:RenderPrice2(option, wnd)
	end
end

function FlipMarketOptions:RenderQuantity( option, wnd )
	if option.eQuantity == FlipMarketOptions.Quantity.CONSTANT then
		wnd:FindChild("ToggleFixedQuantity"):SetText("Constant")
		wnd:FindChild("Quantity"):Show(true)
		wnd:FindChild("Quantity"):SetText(option.nQuantity)
	else
		wnd:FindChild("ToggleFixedQuantity"):SetText("All Except")
		wnd:FindChild("Quantity"):Show(true)
		wnd:FindChild("Quantity"):SetText(option.nQuantity or 0)
	end
end

function FlipMarketOptions:RenderOffset( option, wnd )
	local bOffset = (option.eOffset == FlipMarketOptions.PriceOffset.CONSTANT)

	wnd:FindChild("ToggleDifferenceType"):SetText((bOffset and "Constant") or "Percent")
	wnd:FindChild("ConstantCashDifference"):Show(bOffset)
	if bOffset then wnd:FindChild("ConstantCashDifference"):SetAmount(option.nOffset) end
	wnd:FindChild("PercentageDifference"):Show(not bOffset)
	if not bOffset then wnd:FindChild("PercentageDifference"):SetText(option.nOffset) end
end

function FlipMarketOptions:RenderRelation( option, wnd )
	-- set up the relation type
	local bGreaterThan = (option.eRelation == FlipMarketOptions.PriceComparisons.GREATER)
	wnd:FindChild("ToggleRelation"):SetText((bGreaterThan and "Greater Than") or "Less Than")
end

function FlipMarketOptions:RenderPrice2( option, wnd )
	local strPrice = FlipMarketOptions.PriceOptionsText[option.ePrice2]
	wnd:FindChild("TogglePrice2"):SetText(strPrice)
end

function FlipMarketOptions:OnQuantityChange( wndHandler, wndControl, strText)
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()
	option.nQuantity = tonumber(strText) or 0
end

function FlipMarketOptions:OnConstantCashChange( wndHandler, wndControl, strText)
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()
	option.nOffset = tonumber(strText) or 0
end

function FlipMarketOptions:OnPercentageChange( wndHandler, wndControl, strText)
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()
	option.nOffset = tonumber(strText) or 0
end

function FlipMarketOptions:OnDifferenceTypeToggle( wndHandler )
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()

	local bConstant = option.eOffset == FlipMarketOptions.PriceOffset.CONSTANT

	-- set our option to PERCENT instead
	option.eOffset = (bConstant and FlipMarketOptions.PriceOffset.PERCENT) or FlipMarketOptions.PriceOffset.CONSTANT

	-- set the amount offset to be the amount in the box (don't try to put our offset into the box)
	if bConstant then
		option.nOffset = tonumber(wnd:FindChild("PercentageDifference"):GetText()) or 0
	else
		option.nOffset = wnd:FindChild("ConstantCashDifference"):GetAmount() or 0
	end

	self:RenderOffset(option, wnd)
end

function FlipMarketOptions:OnRelationToggle( wndHandler )
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()

	-- are we less than right now?
	local bLess = option.eRelation == FlipMarketOptions.PriceComparisons.LESS

	-- toggle the stored relation
	option.eRelation = (bLess and FlipMarketOptions.PriceComparisons.GREATER) or FlipMarketOptions.PriceComparisons.LESS

	self:RenderRelation(option, wnd)
end

function FlipMarketOptions:OnPrice2Toggle( wndHandler )
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()

	-- set the new price and the text; can't use # here
	option.ePrice2 = math.max((option.ePrice2 + 1) % 7, 1)
	self:RenderPrice2(option, wnd)
end

-----------------------------------------------------------------------------------------------
-- color options

function FlipMarketOptions:RenderColorOptions( options, wndParent )
	for nKey, option in ipairs(options) do
		local wnd = Apollo.LoadForm(self.xmlDoc, "MarketplaceColorOptions", wndParent, self)
		wnd:FindChild("OptionName"):SetText(option.strName)
		wnd:SetData(option)
	
		local tWindows = {
			GREEN = wnd:FindChild("ColorRowGreen"),
			YELLOW = wnd:FindChild("ColorRowYellow"),
			ORANGE = wnd:FindChild("ColorRowOrange"),
		}
	
		for nKey, wndContainer in pairs(tWindows) do
			wndContainer:SetData(nKey)
			wndContainer:FindChild("ConstantCashDifference"):SetAmount(option.tFixed[nKey] or 0)
			wndContainer:FindChild("PercentageDifference"):SetText(option.tPercent[nKey] or 0)
		end
	end
end

function FlipMarketOptions:OnPercentageDifferenceChanged( wndHandler, wndControl, strText )
	local nKey = wndHandler:GetParent():GetData()
	local option = wndHandler:GetParent():GetParent():GetData()

	option.tPercent[nKey] = tonumber(strText) or 0
end

function FlipMarketOptions:OnConstantCashDifferenceChange( wndHandler, wndControl, strText )
	local nKey = wndHandler:GetParent():GetData()
	local option = wndHandler:GetParent():GetParent():GetData()

	option.tFixed[nKey] = tonumber(strText) or 0
end

-----------------------------------------------------------------------------------------------

function FlipMarketOptions:RenderOptions(eType, wndParent)
	local RenderCategories = {
		BUYOPTION = self.RenderBuyOptions,
		SELLOPTION = self.RenderSellOptions,
		COLOROPTION = self.RenderColorOptions,
	}
	RenderCategories[eType](self, self.options[eType].tOptions, wndParent)
	local wnd = Apollo.LoadForm(self.xmlDoc, "AddNewOption", wndParent, self)
	wndParent:ArrangeChildrenTiles(0)
end

function FlipMarketOptions:AddOption( eType, tOption  )
	local tOptions = self.options[eType].tOptions
	table.insert(tOptions, tOption)
end

function FlipMarketOptions:RemoveOption( eType, tOption )
	local tOptions = self.options[eType].tOptions
	if #tOptions > 1 then
		for nKey, option in ipairs(tOptions) do
			if option == tOption then 
				table.remove(tOptions, nKey)
				return true
			end
		end
	end
	return nil
end

function FlipMarketOptions:SetupInitialOptions()
	local tPriceDifference = {
		strName = "Buy/Sell Price Difference",
		ePrice1 = FlipMarketOptions.Price1Options.BUYTOP1,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.LESS,
		ePrice2 = FlipMarketOptions.Price2Options.SELLTOP1,
	}
	-- self:AddOption("BUYCONDITION", tPriceDifference)
	
	local tMatchBuy = {
		strName = "200 @ Top1 Buy Order Price",
		eQuantity = FlipMarketOptions.Quantity.CONSTANT,
		nQuantity = 200,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.GREATER,
		ePrice2 = FlipMarketOptions.Price2Options.BUYTOP1
	}
	self:AddOption("BUYOPTION", tMatchBuy)
	
	local t1cOvercut = {
		strName = "200 @ 1c Overcut",
		eQuantity = FlipMarketOptions.Quantity.CONSTANT,
		nQuantity = 200,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 1,
		eRelation = FlipMarketOptions.PriceComparisons.GREATER,
		ePrice2 = FlipMarketOptions.Price2Options.BUYTOP1
	}
	self:AddOption("BUYOPTION", t1cOvercut)

	local tAlways = {
		strName = "Always",
		ePrice1 = FlipMarketOptions.Price1Options.ALWAYS,
	}
	-- self:AddOption("SELLCONDITION", tAlways)

	local tNever = {
		strName = "Never",
		ePrice1 = FlipMarketOptions.Price1Options.NEVER,
	}
	-- self:AddOption("SELLCONDITION", tNever)

	local tMatchSell = {
		strName = "All @ Top1 Sell Order Price",
		eQuantity = FlipMarketOptions.Quantity.ALL,
		nQuantity = nil,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 0,
		eRelation = FlipMarketOptions.PriceComparisons.LESS,
		ePrice2 = FlipMarketOptions.Price2Options.SELLTOP1
	}
	self:AddOption("SELLOPTION", tMatchSell)

	local tAll1cUndercut = {
		strName = "All @ 1c Undercut",
		eQuantity = FlipMarketOptions.Quantity.ALL,
		nQuantity = 0,
		eOffset = FlipMarketOptions.PriceOffset.CONSTANT,
		nOffset = 1,
		eRelation = FlipMarketOptions.PriceComparisons.LESS,
		ePrice2 = FlipMarketOptions.Price2Options.SELLTOP1
	}
	self:AddOption("SELLOPTION", tAll1cUndercut)

	local tColors = {
		strName = "Default Color Options",
		tFixed = { GREEN = 10000, YELLOW = 5000, ORANGE = 500 },
		tPercent = { GREEN = 30, YELLOW = 15, ORANGE = 5 },
		tColors = { RED = "red", ORANGE = "xkcdOrangePink", YELLOW = "xkcdAmber", GREEN = "xkcdAlgaeGreen" },
	}
	self:AddOption("COLOROPTION", tColors)
end

-----------------------------------------------------------------------------------------------

function FlipMarketOptions:OnOptionNameChanged( wndHandler, wndControl, strText )
	local wnd = wndHandler:GetParent()
	local option = wnd:GetData()
	option.strName = strText
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

function FlipMarketOption:new(o)
	o = o or {}
	setmetatable(o, self)
	self.__index = self
	o:Init()
	return o
end

-- instance of an "option" is associated with a single group
function FlipMarketOption:Init()
	-- the group 
	self.group = self.group
	
	-- the options object
	self.options = self.options
	
	-- our current selections
	self.active = self.tSave or {}
	
	-- for now, select the first thing in the options
	for eType, value in pairs(self.options) do
		self:Activate(eType, self.active[eType] or 1)
	end
end

function FlipMarketOption:ToSave()
	return self.active
end

-----------------------------------------------------------------------------------------------

function FlipMarketOption:GetActive( eType )
	local tOptions = self.options[eType].tOptions
	if not tOptions[self.active[eType]] then 
		self:Activate(eType, 1) 
	end
	return tOptions[self.active[eType]]
end

function FlipMarketOption:Activate( eType, nKey )
	self.active[eType] = nKey
end

-----------------------------------------------------------------------------------------------

function FlipMarketOption:Display( wndParent )
	for nKey, eType in ipairs(FlipMarketOptions.Ordering) do
		tOption = self.options[eType]
		if not tOption then break end
		
		-- the main window, make it and change the name, set the data to the eType
		local wnd = self.group:AddWindow("GroupOptionsRow", wndParent, self)
		if not wnd then break end
		wnd:FindChild("Title"):SetText(tOption.strDescription)
		wnd:SetData(eType)

		-- fill the dropdown menu
		local wndOptions = wnd:FindChild("Below")
		wndOptions:Show(false)
		wndOptions:SetData(#tOption.tOptions)
		for nKey, tOption in ipairs(tOption.tOptions) do
			local wndOption = self.group:AddWindow("GroupOptionButton", wndOptions, self)
			wndOption:SetData(nKey)
			wndOption:SetText(tOption.strName)
		end
		wndOptions:ArrangeChildrenVert(0)

		-- fill in the "current choice" box
		local btn = wnd:FindChild("GroupOptionButton")
		btn:SetData(false) -- button is not toggled on (dropdown is hidden)
		btn:SetText(self:GetActive(eType).strName or "---inherited---")
	end

	wndParent:ArrangeChildrenVert(0)
end

-----------------------------------------------------------------------------------------------

function FlipMarketOption:OnGroupOptionToggle( wndHandler, wndControl, eMouseButton )
	-- relevant windows
	local wndRow = wndHandler:GetParent():GetParent()
	local wndRight = wndRow:GetParent()
	local wndDropdown = wndRow:FindChild("Below")

	-- is the dropdown open right now?
	local bDropdown = wndHandler:GetData()

	-- the type of option we're selecting
	local eType = wndRow:GetData() 

	-- the number of options
	local nNumOptions = wndDropdown:GetData()

	-- the group
	local group = wndRight:GetData()

	if bDropdown then
		local nLeft, nTop, nRight, nBottom = wndRow:GetAnchorOffsets()
		wndRow:SetAnchorOffsets(nLeft, nTop, nRight, nTop + self.nRowHeight)
		wndHandler:SetData(false)
		wndHandler:SetText(self:GetActive(eType).strName or "---inherited---")
	else 
		self.nRowHeight = self.nRowHeight or wndRow:GetHeight()
		-- offset the height for the main row by nHeight (and a little padding)
		local nLeft, nTop, nRight, nBottom = wndRow:GetAnchorOffsets()
		local nHeight = self.nRowHeight + (nNumOptions * wndHandler:GetHeight()) + 5
		wndRow:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight)

		-- indicate the dropdown is on now, and update the text accordingly
		wndHandler:SetData(true)
		wndHandler:SetText("---")
	end

	-- show the dropdown or turn it off
	wndDropdown:Show(not bDropdown)

	-- rearrange the whole window
	wndRight:ArrangeChildrenVert(0)
end

-----------------------------------------------------------------------------------------------

function FlipMarketOption:OnOptionDropdownChoice( wndHandler, wndControl, eMouseButton )
	-- relevant windows
	local wndRow = wndHandler:GetParent():GetParent()
	
	-- the type of option we're selecting
	local eType = wndRow:GetData() 
	
	-- turn on that option (it's in the button data)
	self:Activate(eType, wndHandler:GetData())
	
	-- pretend we hit the main button again, everything will work nicely
	self:OnGroupOptionToggle(wndRow:FindChild("GroupOptionButton"))
end

-----------------------------------------------------------------------------------------------

function FlipMarketOption:GetColorOptions( nFilterAmount, nBuy, nSell )
	local nBuyPrice = nBuy + (nBuy * MarketplaceLib.kfCommodityBuyOrderTaxMultiplier)
	local nSellPrice = nSell * 0.88
	
	local strColor
	local bSquashed
	
	local options = self:GetActive("COLOROPTION")
	local tFixed = options.tFixed
	local tPercent = options.tPercent
	local tColors = options.tColors
	
	-- if this item hasn't been bought/sold on the marketplace, we're hosed on coloring
	if nSellPrice == 0 or nBuyPrice == 0 then
		strColor = tColors.RED
		bSquashed = nFilterAmount > 0
	else
		-- calculate profit!
		local nDiff = nSellPrice - nBuyPrice
		
		-- this is the percent difference calculation that I chose, deal with it
		local nPercentDiff = (nDiff / ((nBuyPrice + nSellPrice) / 2)) * 100

		-- set the color and figure out if we need to be squashed
		if nDiff >= tFixed.GREEN and nPercentDiff >= tPercent.GREEN then
			strColor = tColors.GREEN
			bSquashed = nFilterAmount > 3
		elseif nDiff >= tFixed.YELLOW and nPercentDiff >= tPercent.YELLOW then
			strColor = tColors.YELLOW
			bSquashed = nFilterAmount > 2
		elseif nDiff >= tFixed.ORANGE and nPercentDiff >= tPercent.ORANGE then
			strColor = tColors.ORANGE
			bSquashed = nFilterAmount > 1
		else
			strColor = tColors.RED
			bSquashed = nFilterAmount > 0
		end
	end

	return options.strName, bSquashed, strColor
end

-- GetBuyOptions : PriceTable -> String Number Number
-- returns the option name and the amount a buy order should be posted at
function FlipMarketOption:GetBuyOptions( tPrice )
	local option = self:GetActive("BUYOPTION")
	local tBase = {
		[FlipMarketOptions.Price2Options.BUYTOP1] = tPrice.nBuy1Price,
		[FlipMarketOptions.Price2Options.BUYTOP10] = tPrice.nBuy10Price,
		[FlipMarketOptions.Price2Options.BUYTOP50] = tPrice.nBuy50Price,
		[FlipMarketOptions.Price2Options.SELLTOP1] = tPrice.nSell1Price,
		[FlipMarketOptions.Price2Options.SELLTOP10] = tPrice.nSell10Price,
		[FlipMarketOptions.Price2Options.SELLTOP50] = tPrice.nSell50Price,
	}
	local nBase = tBase[option.ePrice2] or 0
	local nOffset = option.nOffset
	local nPrice
	if option.eOffset == FlipMarketOptions.PriceOffset.CONSTANT then
		if option.eRelation == FlipMarketOptions.PriceComparisons.LESS then
			-- constant less than b1 price
			nPrice = nBase - nOffset
		else
			-- constant greater than b1 price
			nPrice = nBase + nOffset
		end
	else
		if option.eRelation == FlipMarketOptions.PriceComparisons.LESS then
			-- percentage less than b1 price
			nPrice = nBase - math.max((nBase * (nOffset / 100)), ((nOffset > 0) and 1) or 0)
		else
			-- percentage greater than b1 price
			nPrice = nBase + math.max((nBase * (nOffset / 100)), ((nOffset > 0) and 1) or 0)
		end
	end
	return option.strName, math.max(nPrice, 0), option.nQuantity
end

-- GetSellOptions : PriceTable -> String Number Number
-- returns the option name and the amount a sell order should be posted at
function FlipMarketOption:GetSellOptions( tPrice )
	local option = self:GetActive("SELLOPTION")
	local tBase = {
		[FlipMarketOptions.Price2Options.BUYTOP1] = tPrice.nBuy1Price,
		[FlipMarketOptions.Price2Options.BUYTOP10] = tPrice.nBuy10Price,
		[FlipMarketOptions.Price2Options.BUYTOP50] = tPrice.nBuy50Price,
		[FlipMarketOptions.Price2Options.SELLTOP1] = tPrice.nSell1Price,
		[FlipMarketOptions.Price2Options.SELLTOP10] = tPrice.nSell10Price,
		[FlipMarketOptions.Price2Options.SELLTOP50] = tPrice.nSell50Price,
	}
	local nBase = tBase[option.ePrice2] or 0
	local nOffset = option.nOffset
	local nPrice
	if option.eOffset == FlipMarketOptions.PriceOffset.CONSTANT then
		if option.eRelation == FlipMarketOptions.PriceComparisons.LESS then
			-- constant less than s1 price
			nPrice = nBase - nOffset
		else
			-- constant greater than s1 price
			nPrice = nBase + nOffset
		end
	else
		if option.eRelation == FlipMarketOptions.PriceComparisons.LESS then
			-- percentage less than s1 price
			nPrice = nBase - math.max((nBase * (nOffset / 100)), ((nOffset > 0) and 1) or 0)
		else
			-- percentage greater than s1 price
			nPrice = nBase + math.max((nBase * (nOffset / 100)), ((nOffset > 0) and 1) or 0)
		end
	end
	local nQuantity
	if option.eQuantity == FlipMarketOptions.Quantity.ALLIN then
		nQuantity = option.nQuantity and (-1 * option.nQuantity)
	else
		nQuantity = option.nQuantity
	end
	return option.strName, math.max(nPrice, 0), nQuantity
end

-----------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------

Apollo.RegisterPackage(FlipMarket, "FlipMarket", 1, {})
